"""Controller holding and managing Vantage port devices."""

from aiovantage.models import PortDevice

from .base import BaseController


class PortDevicesController(BaseController[PortDevice]):
    """Controller holding and managing Vantage port devices.

    Port devices are typically "hubs" that communicate with other devices such
    as blinds or lighting systems. It is mostly useful to know about these
    devices so we can set up a proper device hierarchy.
    """

    vantage_types = (
        "Somfy.RS-485_SDN_2_x2E_0_PORT",
        "Somfy.URTSI_2_PORT",
        "Vantage.DmxGateway",
    )
    """The Vantage object types that this controller will fetch."""



"""IConfiguration.OpenFilter method definition."""

from dataclasses import dataclass, field


@dataclass
class OpenFilter:
    """IConfiguration.OpenFilter method definition."""

    interface = "IConfiguration"

    @dataclass
    class Params:
        """Method parameters."""

        object_types: list[str] | None = field(
            default=None,
            metadata={
                "name": "ObjectType",
                "wrapper": "Objects",
            },
        )

        xpath: str | None = field(
            default=None,
            metadata={
                "name": "XPath",
            },
        )

    call: Params | None = field(default=None)
    return_value: int | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""Base class for child device objects."""

from dataclasses import dataclass, field

from .custom_device import CustomDevice
from .types import Parent


@dataclass
class ChildDevice(CustomDevice):
    """Base class for child device objects."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )



"""Base class for all objects."""

from dataclasses import dataclass, field

from xsdata.models.datatype import XmlDateTime


@dataclass(kw_only=True)
class SystemObject:
    """Base class for all objects."""

    id: int = field(
        metadata={
            "name": "VID",
            "type": "Attribute",
        }
    )

    master_id: int = field(
        metadata={
            "name": "Master",
            "type": "Attribute",
        }
    )

    name: str = field(
        metadata={
            "name": "Name",
        }
    )

    model: str = field(
        metadata={
            "name": "Model",
        }
    )

    note: str = field(
        metadata={
            "name": "Note",
        }
    )

    # Not available in 2.x firmware
    mtime: XmlDateTime | None = field(
        default=None,
        metadata={
            "name": "MTime",
            "type": "Attribute",
        },
    )

    # Not available in 2.x firmware
    display_name: str | None = field(
        default=None,
        metadata={
            "name": "DName",
        },
    )

    @property
    def vantage_type(self) -> str:
        """Return the Vantage type of the object."""
        cls = type(self)
        cls_meta = getattr(cls, "Meta", None)
        return getattr(cls_meta, "name", cls.__qualname__)



"""Fetch all thermostats from the Vantage controller, and print any state changes."""

import argparse
import asyncio
import contextlib
import logging
from typing import Any

from aiovantage import Vantage, VantageEvent
from aiovantage.models import Thermostat

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def callback(event: VantageEvent, obj: Thermostat, data: dict[str, Any]) -> None:
    """Print out any state changes."""
    if event == VantageEvent.OBJECT_ADDED:
        print(f"[Thermostat added] '{obj.name}' ({obj.id})")

    elif event == VantageEvent.OBJECT_UPDATED:
        print(f"[Thermostat updated] '{obj.name}' ({obj.id})")
        for attr in data.get("attrs_changed", []):
            print(f"    {attr} = {getattr(obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    async with Vantage(args.host, args.username, args.password) as vantage:
        # Subscribe to updates for all thermostats
        vantage.thermostats.subscribe(callback)

        # Fetch all known thermostats from the controller
        await vantage.thermostats.initialize()

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Base class for system objects in an area."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass(kw_only=True)
class LocationObject(SystemObject):
    """Base class for system objects in an area."""

    # Some objects in firmware 2.x do not have an area_id
    area_id: int | None = field(
        default=None,
        metadata={
            "name": "Area",
        },
    )

    # Some objects in firmware 2.x do not have a location
    location: str | None = field(
        default=None,
        metadata={
            "name": "Location",
        },
    )



"""Interface for querying and controlling thermostats."""

from decimal import Decimal
from enum import IntEnum

from .base import Interface


class ThermostatInterface(Interface):
    """Interface for querying and controlling thermostats."""

    class OperationMode(IntEnum):
        """The operation mode of the thermostat."""

        Off = 0
        Cool = 1
        Heat = 2
        Auto = 3
        Unknown = 4

    class FanMode(IntEnum):
        """The fan mode of the thermostat."""

        Off = 0
        On = 1
        Unknown = 2

    class DayMode(IntEnum):
        """The day mode of the thermostat."""

        Day = 0
        Night = 1
        Unknown = 2
        Standby = 3

    class HoldMode(IntEnum):
        """The hold mode of the thermostat."""

        Normal = 0
        Hold = 1
        Unknown = 2

    class Status(IntEnum):
        """The status of the thermostat."""

        Off = 0
        Cooling = 1
        Heating = 2
        Offline = 3

    method_signatures = {
        "Thermostat.GetIndoorTemperature": Decimal,
        "Thermostat.GetOutdoorTemperature": Decimal,
        "Thermostat.GetHeatSetPoint": Decimal,
        "Thermostat.GetCoolSetPoint": Decimal,
        "Thermostat.GetOperationMode": OperationMode,
        "Thermostat.GetFanMode": FanMode,
        "Thermostat.GetDayMode": DayMode,
        "Thermostat.GetHoldMode": HoldMode,
        "Thermostat.GetStatus": Status,
        "Thermostat.GetAutoSetPoint": Decimal,
    }

    async def get_indoor_temperature(self, vid: int) -> Decimal:
        """Get the current indoor temperature.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The indoor temperature of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetIndoorTemperature
        # -> R:INVOKE <id> <temp> Thermostat.GetIndoorTemperature
        return await self.invoke(
            vid, "Thermostat.GetIndoorTemperature", as_type=Decimal
        )

    async def get_outdoor_temperature(self, vid: int) -> Decimal:
        """Get the current outdoor temperature.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The outdoor temperature of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetOutdoorTemperature
        # -> R:INVOKE <id> <temp> Thermostat.GetOutdoorTemperature
        return await self.invoke(
            vid, "Thermostat.GetOutdoorTemperature", as_type=Decimal
        )

    async def get_heat_set_point(self, vid: int) -> Decimal:
        """Get the current heat set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The heat set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetHeatSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetHeatSetPoint
        return await self.invoke(vid, "Thermostat.GetHeatSetPoint", as_type=Decimal)

    async def set_heat_set_point(self, vid: int, temp: float | Decimal) -> None:
        """Set the current heat set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The heat set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetHeatSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetHeatSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetHeatSetPoint", temp)

    async def get_cool_set_point(self, vid: int) -> Decimal:
        """Get the current cool set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The cool set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetCoolSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetCoolSetPoint
        return await self.invoke(vid, "Thermostat.GetCoolSetPoint", as_type=Decimal)

    async def set_cool_set_point(self, vid: int, temp: float | Decimal) -> None:
        """Set the current cool set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The cool set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetCoolSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetCoolSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetCoolSetPoint", temp)

    async def get_operation_mode(self, vid: int) -> OperationMode:
        """Get the current operation mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The operation mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetOperationMode
        # -> R:INVOKE <id> <mode (Off|Cool|Heat|Auto|Unknown)> Thermostat.GetOperationMode
        return await self.invoke(
            vid, "Thermostat.GetOperationMode", as_type=self.OperationMode
        )

    async def set_operation_mode(self, vid: int, mode: int) -> None:
        """Set the current operation mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The operation mode to set.
        """
        # INVOKE <id> Thermostat.SetOperationMode <mode>
        # -> R:INVOKE <id> Thermostat.SetOperationMode <mode>
        await self.invoke(vid, "Thermostat.SetOperationMode", mode)

    async def get_fan_mode(self, vid: int) -> FanMode:
        """Get the current fan mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The fan mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetFanMode
        # -> R:INVOKE <id> <mode (Off|On|Unknown)> Thermostat.GetFanMode
        return await self.invoke(vid, "Thermostat.GetFanMode", as_type=self.FanMode)

    async def set_fan_mode(self, vid: int, mode: int) -> None:
        """Set the current fan mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The fan mode to set.
        """
        # INVOKE <id> Thermostat.SetFanMode <mode>
        # -> R:INVOKE <id> Thermostat.SetFanMode <mode>
        await self.invoke(vid, "Thermostat.SetFanMode", mode)

    async def get_day_mode(self, vid: int) -> DayMode:
        """Get the current day mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The day mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetDayMode
        # -> R:INVOKE <id> <mode (Day|Night|Unknown|Standby)> Thermostat.GetDayMode
        return await self.invoke(vid, "Thermostat.GetDayMode", as_type=self.DayMode)

    async def set_day_mode(self, vid: int, mode: int) -> None:
        """Set the current day mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The day mode to set.
        """
        # INVOKE <id> Thermostat.SetDayMode <mode>
        # -> R:INVOKE <id> Thermostat.SetDayMode <mode>
        await self.invoke(vid, "Thermostat.SetDayMode", mode)

    async def get_hold_mode(self, vid: int) -> HoldMode:
        """Get the current hold mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The hold mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetHoldMode
        # -> R:INVOKE <id> <mode (Normal|Hold|Unknown)> Thermostat.GetHoldMode
        return await self.invoke(vid, "Thermostat.GetHoldMode", as_type=self.HoldMode)

    async def set_hold_mode(self, vid: int, mode: int) -> None:
        """Set the current hold mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The hold mode to set.
        """
        # INVOKE <id> Thermostat.SetHoldMode <mode>
        # -> R:INVOKE <id> Thermostat.SetHoldMode <mode>
        await self.invoke(vid, "Thermostat.SetHoldMode", mode)

    async def get_status(self, vid: int) -> Status:
        """Get the current status.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The status of the thermostat.
        """
        # INVOKE <id> Thermostat.GetStatus
        # -> R:INVOKE <id> <status (Off|Cooling|Heating|Offline)> Thermostat.GetStatus
        return await self.invoke(vid, "Thermostat.GetStatus", as_type=self.Status)

    async def get_auto_set_point(self, vid: int) -> Decimal:
        """Get the current auto set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The auto set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetAutoSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetAutoSetPoint
        return await self.invoke(vid, "Thermostat.GetAutoSetPoint", as_type=Decimal)

    async def set_auto_set_point(self, vid: int, temp: float | Decimal) -> None:
        """Set the current auto set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The auto set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetAutoSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetAutoSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetAutoSetPoint", temp)



"""Controller holding and managing Vantage light sensors."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.object_interfaces import (
    LightSensorInterface,
    SensorInterface,
)
from aiovantage.models import LightSensor

from .base import BaseController


class LightSensorsController(
    BaseController[LightSensor], LightSensorInterface, SensorInterface
):
    """Controller holding and managing Vantage light sensors."""

    vantage_types = ("LightSensor",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("LIGHT",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a light sensor."""
        state = {
            "level": await LightSensorInterface.get_level(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "LIGHT":
            return

        # STATUS LIGHT
        # -> S:LIGHT <id> <level>
        state = {
            "level": Decimal(args[0]),
        }

        self.update_state(vid, state)



"""Interface for querying and controlling buttons."""

from enum import IntEnum

from .base import Interface


class ButtonInterface(Interface):
    """Interface for querying and controlling buttons."""

    class State(IntEnum):
        """Button state."""

        Up = 0
        Down = 1

    method_signatures = {
        "Button.GetState": State,
    }

    async def get_state(self, vid: int) -> State:
        """Get the state of a button.

        Args:
            vid: The Vantage ID of the button.

        Returns:
            The pressed state of the button, True if pressed, False if not.
        """
        # INVOKE <id> Button.GetState
        # -> R:INVOKE <id> <state (Up/Down)> Button.GetState
        return await self.invoke(vid, "Button.GetState", as_type=self.State)

    async def set_state(self, vid: int, state: State) -> None:
        """Set the state of a button.

        Args:
            vid: The Vantage ID of the button.
            state: The state to set the button to, either a State.Up or State.Down.
        """
        # INVOKE <id> Button.SetState <state (0/1/Up/Down)>
        # -> R:INVOKE <id> <rcode> Button.SetState <state (Up/Down)>
        await self.invoke(vid, "Button.SetState", state)

    # Additional convenience methods, not part of the Vantage API
    async def press(self, vid: int) -> None:
        """Press a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.set_state(vid, self.State.Down)

    async def release(self, vid: int) -> None:
        """Release a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.set_state(vid, self.State.Up)

    async def press_and_release(self, vid: int) -> None:
        """Press and release a button.

        Args:
            vid: The Vantage ID of the button.
        """
        await self.press(vid)
        await self.release(vid)



"""Keypad Station."""

from dataclasses import dataclass, field

from .station_object import StationObject
from .types import Parent


@dataclass
class Keypad(StationObject):
    """Keypad Station."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )



"""Base class for command client interfaces."""

from typing import Any, TypeVar, overload

from aiovantage.command_client import CommandClient
from aiovantage.command_client.utils import (
    ParameterType,
    encode_params,
    parse_param,
    tokenize_response,
)

T = TypeVar("T")


class Interface:
    """Base class for command client object interfaces."""

    method_signatures: dict[str, type[Any] | None] = {}

    def __init__(self, client: CommandClient) -> None:
        """Initialize an object interface for standalone use.

        Args:
            client: The command client to use.
        """
        self._command_client = client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client."""
        return self._command_client

    @overload
    async def invoke(self, vid: int, method: str, *params: ParameterType) -> Any:
        ...

    @overload
    async def invoke(
        self, vid: int, method: str, *params: ParameterType, as_type: type[T]
    ) -> T:
        ...

    async def invoke(
        self,
        vid: int,
        method: str,
        *params: ParameterType,
        as_type: type[T] | None = None,
    ) -> T | Any | None:
        """Invoke a method on an object, and return the parsed response.

        Args:
            vid: The VID of the object to invoke the command on.
            method: The method to invoke.
            params: The parameters to send with the method.
            as_type: The type to cast the response to.

        Returns:
            A parsed response, or None if no response was expected.
        """
        # INVOKE <id> <Interface.Method>
        # -> R:INVOKE <id> <result> <Interface.Method> <arg1> <arg2> ...
        request = f"INVOKE {vid} {method}"
        if params:
            request += f" {encode_params(*params)}"

        # Send the request
        raw_response = await self.command_client.raw_request(request)

        # Break the response into tokens
        _, _, result, _, *args = tokenize_response(raw_response[0])

        # Parse the response
        if as_type is None:
            return self.parse_response(method, result, *args)
        return self.parse_response(method, result, *args, as_type=as_type)

    @overload
    @classmethod
    def parse_response(
        cls, method: str, result: str, *args: str, as_type: type[T]
    ) -> T:
        ...

    @overload
    @classmethod
    def parse_response(cls, method: str, result: str, *args: str) -> Any:
        ...

    @classmethod
    def parse_response(
        cls, method: str, result: str, *args: str, as_type: type[T] | None = None
    ) -> T | Any | None:
        """Parse an object interface "INVOKE" response or status message.

        Args:
            method: The method that was invoked.
            result: The result of the command.
            args: The arguments that were sent with the command.
            as_type: The type to cast the response to.

        Returns:
            A parsed response, or None if no response was expected.
        """
        # -> R:INVOKE <id> <result> <Interface.Method> <arg1> <arg2> ...
        # -> EL: <id> <Interface.Method> <result> <arg1> <arg2> ...
        # -> S:STATUS <id> <Interface.Method> <result> <arg1> <arg2> ...

        # Get the signature of the method we are parsing the response for
        signature = as_type or cls._get_signature(method)

        # Return early if this method has no return value
        if signature is None:
            return None

        # Parse the response
        parsed_response: Any
        if issubclass(signature, tuple) and hasattr(signature, "__annotations__"):
            # If the signature is a NamedTuple, parse each component
            parsed_values: list[Any] = []
            for arg, klass in zip(
                [result, *args], signature.__annotations__.values(), strict=True
            ):
                parsed_values.append(parse_param(arg, klass))

            parsed_response = signature(*parsed_values)
        else:
            # Otherwise, parse a single return value
            parsed_response = parse_param(result, signature)

        # Return the parsed result
        return parsed_response

    @classmethod
    def _get_signature(cls, method: str) -> type[Any] | None:
        # Get the signature of a method.
        for klass in cls.__mro__:
            if issubclass(klass, Interface) and method in klass.method_signatures:
                return klass.method_signatures[method]

        return None



"""Interface for querying and controlling system objects."""

from .base import Interface


class ObjectInterface(Interface):
    """Interface for querying and controlling system objects."""

    method_signatures = {
        "Object.GetMTime": int,
    }

    async def get_mtime(self, vid: int) -> int:
        """Get the modification time of an object.

        Args:
            vid: The Vantage ID of the object.

        Returns:
            The modification time of the object, as a unix timestamp.
        """
        # INVOKE <id> Object.GetMTime
        # -> R:INVOKE <id> <mtime> Object.GetMTime
        return await self.invoke(vid, "Object.GetMTime", as_type=int)



"""Subscribe to events from the Vantage Host Command service."""

import asyncio
import logging
from collections import defaultdict
from collections.abc import Callable, Coroutine, Iterable, Sequence
from contextlib import suppress
from enum import Enum
from inspect import iscoroutinefunction
from ssl import SSLContext
from types import TracebackType
from typing import Literal, TypedDict

from typing_extensions import Self

from aiovantage.errors import ClientConnectionError, ClientError

from .commands import CommandConnection
from .utils import tokenize_response

# The interval between keepalive messages, in seconds.
KEEPALIVE_INTERVAL = 60


class EventType(Enum):
    """Enumeration of event types."""

    CONNECTED = "connect"
    DISCONNECTED = "disconnect"
    RECONNECTED = "reconnect"
    STATUS = "status"
    LOG = "log"
    ENHANCED_LOG = "enhanced_log"


class ConnectEvent(TypedDict):
    """Event emitted when the connection is established."""

    type: Literal[EventType.CONNECTED]


class DisconnectEvent(TypedDict):
    """Event emitted when the connection is lost."""

    type: Literal[EventType.DISCONNECTED]


class ReconnectEvent(TypedDict):
    """Event emitted when the connection is re-established."""

    type: Literal[EventType.RECONNECTED]


class StatusEvent(TypedDict):
    """Event emitted when a "S:" status is received."""

    type: Literal[EventType.STATUS]
    id: int
    status_type: str
    args: Sequence[str]


class EnhancedLogEvent(TypedDict):
    """Event emitted when an "EL:" enhanced log is received."""

    type: Literal[EventType.ENHANCED_LOG]
    log: str


# Type alias for any event type
Event = ConnectEvent | DisconnectEvent | ReconnectEvent | StatusEvent | EnhancedLogEvent

# Type aliases for callbacks for event subscriptions
EventCallback = Callable[[Event], Coroutine | None]
EventFilter = Callable[[Event], bool]
EventSubscription = tuple[EventCallback, EventFilter | None]


class EventStream:
    """Client to subscribe to events from the Vantage Host Command service."""

    def __init__(
        self,
        host: str,
        username: str | None = None,
        password: str | None = None,
        *,
        ssl: SSLContext | bool = True,
        port: int | None = None,
        conn_timeout: float = 30,
    ) -> None:
        """Initialize the client."""
        self._connection = CommandConnection(host, port, ssl, conn_timeout)
        self._username = username
        self._password = password
        self._tasks: list[asyncio.Task[None]] = []
        self._subscriptions: list[EventSubscription] = []
        self._status_subscribers: dict[str, int] = defaultdict(int)
        self._enhanced_log_subscribers: dict[str, int] = defaultdict(int)
        self._start_lock = asyncio.Lock()
        self._started = False
        self._connection_lock = asyncio.Lock()
        self._command_queue: asyncio.Queue[str] = asyncio.Queue()
        self._logger = logging.getLogger(__name__)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit context manager."""
        self.stop()
        if exc_val:
            raise exc_val

    async def start(self) -> None:
        """Initialize the event stream."""
        async with self._start_lock:
            if self._started:
                return

            await self.get_connection()
            self._tasks.append(asyncio.create_task(self._message_handler()))
            self._tasks.append(asyncio.create_task(self._command_handler()))
            self._tasks.append(asyncio.create_task(self._keepalive()))

            self._logger.debug("Started the event stream")
            self._started = True

    def stop(self) -> None:
        """Stop the event stream."""
        for task in self._tasks:
            task.cancel()
        self._tasks.clear()
        self._connection.close()

        self._logger.debug("Stopped the event stream")
        self._started = False

    async def get_connection(self) -> CommandConnection:
        """Get a connection to the Host Command service."""
        async with self._connection_lock:
            if self._connection.closed:
                # Open a new connection
                await self._connection.open()

                # Authenticate the new connection if we have credentials
                if self._username and self._password:
                    await self._send(f"LOGIN {self._username} {self._password}")

                self._logger.info(
                    "Connected to event stream at %s:%d",
                    self._connection.host,
                    self._connection.port,
                )

            return self._connection

    def subscribe(
        self,
        callback: EventCallback,
        event_filter: EventType | Iterable[EventType] | EventFilter | None = None,
    ) -> Callable[[], None]:
        """Subscribe to events from the Host Command service.

        Args:
            callback: The callback to invoke when an event is received.
            event_filter: A filter to apply to events before invoking the callback.

        Returns:
            A function that can be used to unsubscribe from events.
        """
        # Support filtering by event type, a list of event types, or a predicate
        subscription: EventSubscription
        if isinstance(event_filter, EventType):
            subscription = (callback, lambda event: event["type"] == event_filter)
        elif isinstance(event_filter, Iterable):
            subscription = (callback, lambda event: event["type"] in event_filter)  # type: ignore[operator]
        else:
            subscription = (callback, event_filter)

        self._subscriptions.append(subscription)

        # Return an unsubscribe callback to remove the subscription
        def unsubscribe() -> None:
            self._subscriptions.remove(subscription)

        return unsubscribe

    def subscribe_status(
        self, callback: EventCallback, status_types: str | Iterable[str]
    ) -> Callable[[], None]:
        """Subscribe to "Status" events from the Host Command service.

        Args:
            callback: The callback to invoke when an event is received.
            status_types: The status types to subscribe to status events for.

        Returns:
            A coroutine that can be used to unsubscribe from status events.
        """
        # Support both a single status type and a list of status types
        if isinstance(status_types, str):
            status_types = (status_types,)

        # Enable this status type if it's not already enabled
        for status_type in status_types:
            self._status_subscribers[status_type] += 1
            if self._status_subscribers[status_type] == 1:
                self._enable_status(status_type)

        # Subscribe, and return an unsubscribe callback
        remove_subscription = self.subscribe(
            callback,
            lambda event: (
                event["type"] == EventType.STATUS
                and event["status_type"] in status_types
            ),
        )

        def unsubscribe() -> None:
            for status_type in status_types:
                self._status_subscribers[status_type] -= 1
                if self._status_subscribers[status_type] == 0:
                    self._disable_status(status_type)
            remove_subscription()

        return unsubscribe

    def subscribe_enhanced_log(
        self, callback: EventCallback, log_types: str | Iterable[str]
    ) -> Callable[[], None]:
        """Subscribe to "Enhanced Log" events from the Host Command service.

        Args:
            callback: The callback to invoke when an event is received.
            log_types: The event log type or types to subscribe to.

        Returns:
            A coroutine that can be used to unsubscribe from log events.
        """
        if isinstance(log_types, str):
            log_types = (log_types,)

        # Enable this log type if it's not already enabled
        for log_type in log_types:
            self._enhanced_log_subscribers[log_type] += 1
            if self._enhanced_log_subscribers[log_type] == 1:
                self._enable_enhanced_log(log_type)

        # Subscribe, and return an unsubscribe callback
        remove_subscription = self.subscribe(callback, EventType.ENHANCED_LOG)

        def unsubscribe() -> None:
            for log_type in log_types:
                self._enhanced_log_subscribers[log_type] -= 1
                if self._enhanced_log_subscribers[log_type] == 0:
                    self._disable_enhanced_log(log_type)
            remove_subscription()

        return unsubscribe

    def emit(self, event: Event) -> None:
        """Emit an event to subscribers.

        Args:
            event: The event to emit.
        """
        for callback, event_filter in self._subscriptions:
            if event_filter is None or event_filter(event):
                if iscoroutinefunction(callback):
                    asyncio.create_task(callback(event))
                else:
                    callback(event)

    async def _message_handler(self) -> None:
        # Handle incoming messages from the Host Command service.
        connect_attempts = 0
        while True:
            connect_attempts += 1
            try:
                # Get the connection to the Host Command service
                conn = await self.get_connection()

                # Notify that we're connected
                if connect_attempts == 1:
                    self.emit({"type": EventType.CONNECTED})
                else:
                    self.emit({"type": EventType.RECONNECTED})
                    self._resubscribe()
                connect_attempts = 1

                # Wait for new messages
                while True:
                    message = await conn.readuntil(b"\r\n")
                    message = message.rstrip()
                    self._logger.debug("Received message: %s", message)
                    self._parse_message(message)

            except ClientConnectionError:
                pass  # Pass through to retry logic below

            # If we get here, the connection was lost
            self.emit({"type": EventType.DISCONNECTED})

            # Clear the command queue
            with suppress(asyncio.QueueEmpty):
                while not self._command_queue.empty():
                    self._command_queue.get_nowait()
                    self._command_queue.task_done()

            # Attempt to reconnect, with backoff
            reconnect_wait = min(2 * connect_attempts, 600)
            self._logger.debug(
                "Disconnected from EventStream - retrying in %d seconds",
                reconnect_wait,
            )

            # Log a warning every 10 attempts
            if connect_attempts % 10 == 0:
                self._logger.warning(
                    "%d attempts to (re)connect to controller failed"
                    " - This might be an indication of connection issues.",
                    connect_attempts,
                )

            await asyncio.sleep(reconnect_wait)

    async def _command_handler(self) -> None:
        # Handle outgoing commands to the Host Command service.
        while True:
            try:
                command = await self._command_queue.get()
                await self._send(command)
                self._command_queue.task_done()
            except ClientError as err:
                self._logger.warning("Error while sending command: %s", str(err))

    async def _keepalive(self) -> None:
        # Send a periodic "ECHO" keepalive command.
        while True:
            await asyncio.sleep(KEEPALIVE_INTERVAL)

            try:
                await self._send("ECHO")
            except ClientError as err:
                self._logger.debug("Error while sending keepalive: %s", str(err))

    async def _send(self, message: str) -> None:
        # Send a plaintext message to the Host Command service."""
        self._logger.debug("Sending message: %s", message)
        await self._connection.write(f"{message}\n")

    def _parse_message(self, message: str) -> None:
        # Parse a message from the Host Command service.
        if message.startswith("S:"):
            # Parse a "status" message, of the form "S:<type> <id> <args>"
            # These messages are emitted when the state of an object changes after
            # subscribing to updates via "STATUS <type>" or "ADDSTATUS <vid>".
            status_type, id_str, *args = tokenize_response(message)
            self.emit(
                {
                    "type": EventType.STATUS,
                    "status_type": status_type[2:],
                    "id": int(id_str),
                    "args": args,
                }
            )
        elif message.startswith("EL:"):
            # Parse an "enhanced log" message, of the form "EL: <log>"
            # These messages are emitted when an enhanced log is received after
            # subscribing to updates via "ELLOG <type>".
            self.emit({"type": EventType.ENHANCED_LOG, "log": message[4:]})
        elif message.startswith("R:ERROR"):
            self._logger.error("Error message from EventStream: %s", message)

    def _queue_command(self, command: str) -> None:
        # Queue a command to be sent to the Host Command service.
        self._command_queue.put_nowait(command)

    def _enable_status(self, status_type: str) -> None:
        # Enable status updates on the controller for a particular status type.
        self._queue_command(f"STATUS {status_type}")

    def _disable_status(self, _status_type: str) -> None:
        # Disable status updates on the controller for a particular status type.
        # Note: This assumes the count has already been decremented.
        self._queue_command("STATUS NONE")
        for status_type, count in self._status_subscribers.items():
            if count > 0:
                self._queue_command(f"STATUS {status_type}")

    def _enable_enhanced_log(self, log_type: str) -> None:
        # Enable enhanced logging on the controller for a particular log type.
        self._queue_command("ELAGG 1 ON")
        self._queue_command(f"ELENABLE {log_type} ON")
        self._queue_command(f"ELLOG {log_type} ON")

    def _disable_enhanced_log(self, log_type: str) -> None:
        # Disable enhanced logging on the controller for a particular log type.
        self._queue_command(f"ELLOG {log_type} OFF")

    def _resubscribe(self) -> None:
        # Re-subscribe to events after a reconnection.
        for status_type, count in self._status_subscribers.items():
            if count > 0:
                self._enable_status(status_type)

        for log_type, count in self._enhanced_log_subscribers.items():
            if count > 0:
                self._enable_enhanced_log(log_type)



"""Somfy RS-485 SDN 2.0 blind group."""

from dataclasses import dataclass

from .blind_group_base import BlindGroupBase
from .child_device import ChildDevice


@dataclass
class SomfyRS485GroupChild(BlindGroupBase, ChildDevice):
    """Somfy RS-485 SDN 2.0 blind group."""

    class Meta:
        name = "Somfy.RS-485_Group_CHILD"



"""Interface for querying and controlling sensors."""

from decimal import Decimal

from .base import Interface


class SensorInterface(Interface):
    """Interface for querying and controlling sensors."""

    method_signatures = {
        "Sensor.GetLevel": Decimal,
        "Sensor.GetLevelHW": Decimal,
    }

    async def get_level(self, vid: int) -> Decimal:
        """Get the level of a sensor, using cached value if available.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The level of the sensor.
        """
        # INVOKE <id> Sensor.GetLevel
        # -> R:INVOKE <id> <level (0-100)> Sensor.GetLevel
        return await self.invoke(vid, "Sensor.GetLevel", as_type=Decimal)

    async def get_level_hw(self, vid: int) -> Decimal:
        """Get the level of a sensor directly from the hardware.

        Args:
            vid: The Vantage ID of the sensor.

        Returns:
            The level of the sensor.
        """
        # INVOKE <id> Sensor.GetLevelHW
        # -> R:INVOKE <id> <level (0-100)> Sensor.GetLevelHW
        return await self.invoke(vid, "Sensor.GetLevelHW", as_type=Decimal)



"""Controller holding and managing Vantage tasks."""

from typing_extensions import override

from aiovantage.command_client.object_interfaces import TaskInterface
from aiovantage.models import Task

from .base import BaseController


class TasksController(BaseController[Task], TaskInterface):
    """Controller holding and managing Vantage tasks."""

    vantage_types = ("Task",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("TASK",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    interface_status_types = ("Task.IsRunning",)
    """Which object interface status messages this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a task."""
        state = {
            "is_running": await TaskInterface.is_running(self, vid),
            "state": await TaskInterface.get_state(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "TASK":
            return

        # STATUS TASK
        # -> S:TASK <id> <state>
        state = {
            "state": int(args[0]),
        }

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        if method != "Task.IsRunning":
            return

        state = {
            "is_running": self.parse_response(method, result, *args),
        }

        self.update_state(vid, state)



"""Base class for Vantage port devices."""


from dataclasses import dataclass

from .location_object import LocationObject

# NOTE: Inherits from LocationObject on 2.x firmware, ParentDevice on 3.x firmware.


@dataclass
class PortDevice(LocationObject):
    """Base class for Vantage port devices."""



"""Base controller for Vantage objects."""

import asyncio
import logging
from collections.abc import Callable, Iterable
from dataclasses import fields
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from aiovantage.command_client import CommandClient, Event, EventStream, EventType
from aiovantage.command_client.utils import tokenize_response
from aiovantage.config_client import ConfigClient
from aiovantage.config_client.requests import get_objects
from aiovantage.events import EventCallback, VantageEvent
from aiovantage.models import SystemObject
from aiovantage.query import QuerySet

if TYPE_CHECKING:
    from aiovantage import Vantage

T = TypeVar("T", bound=SystemObject)


# Types for state and subscriptions
EventSubscription = tuple[EventCallback[T], Iterable[VantageEvent] | None]


class BaseController(QuerySet[T]):
    """Base controller for Vantage objects."""

    vantage_types: tuple[str, ...]
    """The Vantage object types that this controller handles."""

    status_types: tuple[str, ...] | None = None
    """Which Vantage 'STATUS' types this controller handles, if any."""

    interface_status_types: tuple[str, ...] | Literal["*"] | None = None
    """Which object interface status messages this controller handles, if any."""

    def __init__(self, vantage: "Vantage") -> None:
        """Initialize a controller.

        Args:
            vantage: The Vantage instance.
        """
        self._vantage = vantage
        self._items: dict[int, T] = {}
        self._logger = logging.getLogger(__package__)
        self._subscribed_to_state_changes = False
        self._subscriptions: list[EventSubscription[T]] = []
        self._id_subscriptions: dict[int, list[EventSubscription[T]]] = {}
        self._initialized = False
        self._lock = asyncio.Lock()

        QuerySet.__init__(self, self._items, self._lazy_initialize)

        self.__post_init__()

    def __post_init__(self) -> None:
        """Post initialization hook for subclasses."""

    def __getitem__(self, vid: int) -> T:
        """Return the object with the given Vantage ID."""
        return self._items[vid]

    def __contains__(self, vid: int) -> bool:
        """Return True if the object with the given Vantage ID exists."""
        return vid in self._items

    @property
    def config_client(self) -> ConfigClient:
        """Return the config client instance."""
        return self._vantage.config_client

    @property
    def command_client(self) -> CommandClient:
        """Return the command client instance."""
        return self._vantage.command_client

    @property
    def event_stream(self) -> EventStream:
        """Return the event stream instance."""
        return self._vantage.event_stream

    @property
    def initialized(self) -> bool:
        """Return True if this controller has been initialized."""
        return self._initialized

    @property
    def stateful(self) -> bool:
        """Return True if this controller manages stateful objects."""
        return bool(self.status_types or self.interface_status_types)

    @property
    def known_ids(self) -> set[int]:
        """Return a set of all known object IDs."""
        return set(self._items.keys())

    async def fetch_object_state(self, _vid: int) -> None:
        """Fetch the full state of an object.

        Should be overridden by subclasses that manage stateful objects.
        """
        return

    def handle_status(self, _vid: int, _status: str, *_args: str) -> None:
        """Handle simple status messages from the event stream.

        Should be overridden by subclasses that manage stateful objects using
        "STATUS {type}" messages.
        """
        return

    def handle_interface_status(
        self, _vid: int, _method: str, _result: str, *_args: str
    ) -> None:
        """Handle object interface status messages from the event stream.

        Should be overridden by subclasses that manage stateful objects using object
        interface status messages from "ADDSTATUS {vid}" or "ELLOG STATUS" events.
        """
        return

    async def initialize(self, fetch_state: bool = True) -> None:
        """Populate the controller, and optionally fetch initial state.

        Args:
            fetch_state: Whether to also fetch the state of each object.
        """
        # Prevent concurrent controller initialization from multiple tasks, since we
        # are batch-modifying the _items dict.
        async with self._lock:
            prev_ids = set(self._items.keys())
            cur_ids = set()

            # Fetch all objects managed by this controller
            async for obj in get_objects(self.config_client, types=self.vantage_types):
                if obj.id in prev_ids:
                    # This is an existing object.
                    # Update any attributes that have changed and notify subscribers.
                    # Ignore the mtime attribute, and any state attributes.
                    self.update_state(
                        obj.id,
                        {
                            field.name: getattr(obj, field.name)
                            for field in fields(type(obj))
                            if field.name != "mtime"
                            and field.metadata.get("type") != "Ignore"
                        },
                    )
                else:
                    # This is a new object.
                    # Add it to the controller and notify subscribers
                    self._items[obj.id] = obj
                    self.emit(VantageEvent.OBJECT_ADDED, obj)

                    # Fetch the state of stateful objects
                    if self.stateful and fetch_state:
                        await self.fetch_object_state(obj.id)

                # Keep track of which objects we've seen
                cur_ids.add(obj.id)

            # Handle objects that were removed
            for vid in prev_ids - cur_ids:
                obj = self._items.pop(vid)
                self.emit(VantageEvent.OBJECT_DELETED, obj)

        # Subscribe to state changes for objects managed by this controller
        if fetch_state and len(self._items) > 0:
            await self.subscribe_to_state_changes()

        # Mark the controller as initialized
        if not self._initialized:
            self._initialized = True

        self._logger.info(
            "%s initialized (%d objects)", type(self).__name__, len(self._items)
        )

    async def fetch_full_state(self) -> None:
        """Fetch the full state of all objects managed by this controller."""
        if not self.stateful:
            return

        for obj in self._items.values():
            await self.fetch_object_state(obj.id)

        self._logger.info("%s fetched state", type(self).__name__)

    async def subscribe_to_state_changes(self) -> None:
        """Subscribe to state changes for objects managed by this controller."""
        if self._subscribed_to_state_changes or not self.stateful:
            return

        # Ensure that the event stream is running
        await self.event_stream.start()

        # Subscribe to "STATUS {type}" updates, if this controller cares about them.
        if self.status_types:
            self.event_stream.subscribe_status(self._handle_event, self.status_types)

        # Some state changes are only available from "object" status events.
        # These can be subscribed to by using "STATUSADD {vid}" or "ELLOG STATUS".
        if self.interface_status_types:
            # Subscribe to "object status" events from the Enhanced Log.
            self.event_stream.subscribe_enhanced_log(
                self._handle_event, ("STATUS", "STATUSEX")
            )

        self._subscribed_to_state_changes = True
        self._logger.info("%s subscribed to state changes", type(self).__name__)

    def subscribe(
        self,
        callback: EventCallback[T],
        id_filter: int | Iterable[int] | None = None,
        event_filter: VantageEvent | Iterable[VantageEvent] | None = None,
    ) -> Callable[[], None]:
        """Subscribe to status changes for objects managed by this controller.

        Args:
            callback: The callback to call when an object changes.
            id_filter: The Vantage IDs to subscribe to, all objects if None.
            event_filter: The event types to subscribe to, all events if None.

        Returns:
            A function to unsubscribe from the callback.
        """
        # Handle single ID filter or single event filter
        if isinstance(id_filter, int):
            id_filter = (id_filter,)

        if isinstance(event_filter, VantageEvent):
            event_filter = (event_filter,)

        # Create the subscription
        subscription = (callback, event_filter)

        # Add the subscription to the list of subscriptions
        if id_filter is None:
            self._subscriptions.append(subscription)
        else:
            for vid in id_filter:
                if vid not in self._id_subscriptions:
                    self._id_subscriptions[vid] = []
                self._id_subscriptions[vid].append(subscription)

        # Return a function to unsubscribe
        def unsubscribe() -> None:
            if id_filter is None:
                self._subscriptions.remove(subscription)
            else:
                for vid in id_filter:
                    if vid not in self._id_subscriptions:
                        continue
                    self._id_subscriptions[vid].remove(subscription)

        return unsubscribe

    def emit(
        self, event_type: VantageEvent, obj: T, data: dict[str, Any] | None = None
    ) -> None:
        """Emit an event to subscribers of this controller.

        Args:
            event_type: The type of event to emit.
            obj: The object that the event relates to.
            data: Data to pass to the callback.
        """
        if data is None:
            data = {}

        # Grab a list of subscribers that care about this object
        subscribers = self._subscriptions + self._id_subscriptions.get(obj.id, [])
        for callback, event_filter in subscribers:
            if event_filter is not None and event_type not in event_filter:
                continue

            if iscoroutinefunction(callback):
                asyncio.create_task(callback(event_type, obj, data))  # noqa: RUF006
            else:
                callback(event_type, obj, data)

    def update_state(self, vid: int, attrs: dict[str, Any]) -> None:
        """Update the attributes of an object and notify subscribers of changes."""
        # Ignore updates for objects that this controller doesn't manage
        if (obj := self._items.get(vid)) is None:
            return

        # Check if any state attributes changed and update them
        attrs_changed = []
        for key, value in attrs.items():
            try:
                if getattr(obj, key) != value:
                    setattr(obj, key, value)
                    attrs_changed.append(key)
            except AttributeError:
                self._logger.warning("Object '%d' has no attribute '%s'", obj.id, key)

        # Notify subscribers if any attributes changed
        if len(attrs_changed) > 0:
            self.emit(
                VantageEvent.OBJECT_UPDATED,
                obj,
                {"attrs_changed": attrs_changed},
            )

    async def _handle_event(self, event: Event) -> None:
        # Handle events from the event stream
        if event["type"] == EventType.STATUS:
            # Ignore events for objects that this controller doesn't manage
            if event["id"] not in self._items:
                return

            if event["status_type"] == "STATUS":
                # Handle "object interface" status events of the form:
                # -> S:STATUS <id> <method> <result> <arg1> <arg2> ...
                method, result, *args = event["args"]
                self.handle_interface_status(event["id"], method, result, *args)
            else:
                # Handle "category" status events, eg: S:LOAD, S:BLIND, etc
                self.handle_status(event["id"], event["status_type"], *event["args"])

        elif event["type"] == EventType.ENHANCED_LOG:
            # We only ever subscribe to STATUS/STATUSEX logs from the enhanced log.
            # These are "object interface" status messages, of the form:
            #   EL: <id> <method> <result> <arg1> <arg2> ...
            vid_str, method, result, *args = tokenize_response(event["log"])
            vid = int(vid_str)

            # Ignore events for objects that this controller doesn't manage
            if vid not in self._items:
                return

            # Pass the event to the controller
            self.handle_interface_status(vid, method, result, *args)

    async def _lazy_initialize(self) -> None:
        # Initialize the controller if it isn't already initialized
        if not self._initialized:
            await self.initialize()



"""Interface for querying and controlling tasks."""

from .base import Interface


class TaskInterface(Interface):
    """Interface for querying and controlling tasks."""

    method_signatures = {
        "Task.IsRunning": bool,
        "Task.GetState": int,
    }

    async def start(self, vid: int) -> None:
        """Start a task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.Start <source> <event> <param1> <param2>
        # -> R:INVOKE <id> <rcode (0/1)> Task.Start <source> <event> <param1> <param2>
        await self.invoke(vid, "Task.Start")

    async def stop(self, vid: int) -> None:
        """Stop a running task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.Stop
        # -> R:INVOKE <id> <rcode> Task.Stop
        await self.invoke(vid, "Task.Stop")

    async def cancel(self, vid: int) -> None:
        """Cancel a scheduled task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.Cancel
        # -> R:INVOKE <id> <rcode> Task.Cancel
        await self.invoke(vid, "Task.Cancel")

    async def is_running(self, vid: int) -> bool:
        """Get the running state of a task.

        Args:
            vid: The Vantage ID of the task.
        """
        # INVOKE <id> Task.IsRunning
        # -> R:INVOKE <id> <running (0/1)> Task.IsRunning
        return await self.invoke(vid, "Task.IsRunning", as_type=bool)

    async def get_state(self, vid: int) -> int:
        """Get the state of a task.

        Args:
            vid: The Vantage ID of the task.

        Returns:
            The LED state of the task.
        """
        # INVOKE <id> Task.GetState
        # -> R:INVOKE <id> <state> Task.GetState
        return await self.invoke(vid, "Task.GetState", as_type=int)

    async def set_state(self, vid: int, state: int) -> None:
        """Set the state of a task.

        Args:
            vid: The Vantage ID of the task.
            state: The state to set the task to.
        """
        # INVOKE <id> Task.SetState <state>
        # -> R:INVOKE <id> <rcode> Task.SetState <state>
        await self.invoke(vid, "Task.SetState", state)



"""Introspection interface (IIntrospection) method definitions."""

from .get_interfaces import GetInterfaces
from .get_sys_info import GetSysInfo
from .get_types import GetTypes
from .get_version import GetVersion

__all__ = [
    "GetInterfaces",
    "GetSysInfo",
    "GetTypes",
    "GetVersion",
]



"""Helper utilities for discovering details about Vantage controllers."""
import re
from dataclasses import dataclass

from .command_client.commands import CommandClient
from .config_client import ConfigClient
from .config_client.interfaces.introspection.get_sys_info import GetSysInfo
from .errors import (
    ClientConnectionError,
    ClientError,
    LoginFailedError,
    LoginRequiredError,
)


@dataclass
class VantageControllerDetails:
    """Details about a queried Vantage controller."""

    host: str
    supports_ssl: bool
    requires_auth: bool


async def get_controller_details(host: str) -> VantageControllerDetails | None:
    """Discover Vantage controller details from given hostname/ip.

    Args:
        host: The hostname/ip of a Vantage controller.

    Returns:
        A DiscoveredVantageController, or None if a controller could not be reached.
    """
    try:
        # Try connecting with SSL first
        async with CommandClient(host) as client:
            supports_ssl = True
            requires_auth = await _auth_required(client)
    except ClientConnectionError:
        # If SSL fails, try connecting without SSL
        try:
            async with CommandClient(host, ssl=False) as client:
                supports_ssl = False
                requires_auth = await _auth_required(client)
        except ClientConnectionError:
            # If both fail, the controller is unreachable
            return None

    return VantageControllerDetails(host, supports_ssl, requires_auth)


async def is_auth_required(host: str, ssl: bool = True) -> bool:
    """Check if authentication is required for the given controller.

    Args:
        host: The hostname/ip of the Vantage controller.
        ssl: Whether to use SSL when connecting to the controller.

    Returns:
        True if authentication is required, False otherwise.

    Raises:
        ClientConnectionError: If a controller could not be reached.
    """
    async with CommandClient(host, ssl=ssl) as client:
        return await _auth_required(client)


async def validate_credentials(
    host: str, username: str, password: str, ssl: bool = True
) -> bool:
    """Check if the given credentials are valid for the given controller.

    Args:
        host: The hostname/ip of the Vantage controller.
        ssl: Whether to use SSL when connecting to the controller.
        username: The username to check.
        password: The password to check.

    Returns:
        True if the credentials are valid, False otherwise.

    Raises:
        ClientConnectionError: If a controller could not be reached.
    """
    try:
        async with CommandClient(host, username, password, ssl=ssl) as client:
            await client.command("VERSION")
            return True
    except LoginFailedError:
        return False


async def get_serial_from_controller(
    host: str,
    username: str | None = None,
    password: str | None = None,
    ssl: bool = True,
) -> int | None:
    """Get the serial number of the given controller.

    Args:
        host: The hostname/ip of the Vantage controller.
        username: The username to check, or None if not required.
        password: The password to check, or None if not required.
        ssl: Whether to use SSL when connecting to the controller.

    Returns:
        The serial number of the controller, or None if not found.
    """
    try:
        async with ConfigClient(host, username, password, ssl=ssl) as client:
            sys_info_response = await client.request(GetSysInfo)
            if sys_info_response is None:
                return None

            return sys_info_response.sys_info.serial_number

    except ClientError:
        return None


def get_serial_from_hostname(hostname: str) -> str | None:
    """Get the serial number from a Vantage mDNS hostname.

    Args:
        hostname: The hostname to parse.

    Returns:
        The serial number of the controller, or None if not found.
    """
    match = re.match(r"ic-ii-(?P<serial_number>\d+)", hostname)
    if not match:
        return None
    return match.group("serial_number")


async def _auth_required(client: CommandClient) -> bool:
    # Check if authentication is required for the given client.
    try:
        await client.command("VERSION")
        return False
    except LoginRequiredError:
        return True



"""Interface for querying and controlling blinds."""

from decimal import Decimal
from typing import NamedTuple

from .base import Interface


class BlindInterface(Interface):
    """Interface for querying and controlling blinds."""

    class BlindState(NamedTuple):
        """The state of a blind."""

        is_moving: bool
        """Is the blind currently moving?"""

        start_pos: Decimal
        """Position the blind is moving from (as a percentage)"""

        end_pos: Decimal
        """Position the blind is moving to (as a percentage)"""

        transition_time: Decimal
        """Time the blind will take to move (in seconds)"""

        start_time: int
        """Time the blind started moving (in milliseconds since start of UTC day)"""

    method_signatures = {
        "Blind.GetPosition": Decimal,
        "Blind.GetPositionHW": Decimal,
        "Blind.GetTiltAngle": int,
        "Blind.GetTiltAngleHW": int,
        "Blind.IsTiltAvailable": bool,
        "Blind.GetBlindState": BlindState,
    }

    async def open(self, vid: int) -> None:
        """Open a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Open
        # -> R:INVOKE <id> <rcode> Blind.Open
        await self.invoke(vid, "Blind.Open")

    async def close(self, vid: int) -> None:
        """Close a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Close
        # -> R:INVOKE <id> <rcode> Blind.Close
        await self.invoke(vid, "Blind.Close")

    async def stop(self, vid: int) -> None:
        """Stop a blind.

        Args:
            vid: The Vantage ID of the blind.
        """
        # INVOKE <id> Blind.Stop
        # -> R:INVOKE <id> <rcode> Blind.Stop
        await self.invoke(vid, "Blind.Stop")

    async def set_position(self, vid: int, position: float) -> None:
        """Set the position of a blind.

        Args:
            vid: The Vantage ID of the blind.
            position: The position to set the blind to, as a percentage.
        """
        # INVOKE <id> Blind.SetPosition <position>
        # -> R:INVOKE <id> <rcode> Blind.SetPosition <position>
        await self.invoke(vid, "Blind.SetPosition", position)

    async def get_position(self, vid: int) -> Decimal:
        """Get the position of a blind, using cached value if available.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The position of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetPosition
        # -> R:INVOKE <id> <position (0-100.000)> Blind.GetPosition
        return await self.invoke(vid, "Blind.GetPosition", as_type=Decimal)

    async def get_position_hw(self, vid: int) -> Decimal:
        """Get the position of a blind directly from the hardware.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The position of the blind, as a percentage.
        """
        # INVOKE <id> Blind.GetPositionHW
        # -> R:INVOKE <id> <position (0-100.000)> Blind.GetPositionHW
        return await self.invoke(vid, "Blind.GetPositionHW", as_type=Decimal)

    # Methods below here are not available in 2.x firmware.
    async def set_tilt_angle(self, vid: int, angle: int) -> None:
        """Set the tilt angle of a blind.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle to set the blind to, from -100 to 100.
        """
        # INVOKE <id> Blind.SetTiltAngle <angle>
        # -> R:INVOKE <id> <rcode> Blind.SetTiltAngle <angle>
        await self.invoke(vid, "Blind.SetTiltAngle", angle)

    async def get_tilt_angle(self, vid: int) -> int:
        """Get the tilt angle of a blind, using cached value if available.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The tilt angle of the blind, from -100 to 100.
        """
        # INVOKE <id> Blind.GetTiltAngle
        # -> R:INVOKE <id> <angle (-100-100)> Blind.GetTiltAngle
        return await self.invoke(vid, "Blind.GetTiltAngle", as_type=int)

    async def get_tilt_angle_hw(self, vid: int) -> int:
        """Get the tilt angle of a blind directly from the hardware.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The tilt angle of the blind, from -100 to 100.
        """
        # INVOKE <id> Blind.GetTiltAngleHW
        # -> R:INVOKE <id> <angle (-100-100)> Blind.GetTiltAngleHW
        return await self.invoke(vid, "Blind.GetTiltAngleHW", as_type=int)

    async def tilt_clockwise(self, vid: int, angle: int) -> None:
        """Tilt the blinds clockwise by the specified angle.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltClockwise <angle>
        await self.invoke(vid, "Blind.TiltClockwise", angle)

    async def tilt_counter_clockwise(self, vid: int, angle: int) -> None:
        """Tilt the shades counter-clockwise by the specified angle.

        Args:
            vid: The Vantage ID of the blind.
            angle: The angle offset the blinds should be tilted.
        """
        # INVOKE <id> Blind.TiltCounterClockwise <angle>
        # -> R:INVOKE <id> <rcode> Blind.TiltCounterClockwise <angle>
        await self.invoke(vid, "Blind.TiltCounterClockwise", angle)

    async def is_tilt_available(self, vid: int) -> bool:
        """Check if the blind can tilt in its current state.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            Whether the blind supports tilting.
        """
        # INVOKE <id> Blind.IsTiltAvailable
        # -> R:INVOKE <id> <available (0/1)> Blind.IsTiltAvailable
        return await self.invoke(vid, "Blind.IsTiltAvailable", as_type=bool)

    async def get_blind_state(self, vid: int) -> BlindState:
        """Get the state of a blind.

        Args:
            vid: The Vantage ID of the blind.

        Returns:
            The state of the blind.
        """
        # INVOKE <id> Blind.GetBlindState
        # -> R:INVOKE <id> <moving> Blind.GetBlindState <start> <end> <transitionTime> <startTime>
        return await self.invoke(vid, "Blind.GetBlindState", as_type=self.BlindState)



"""Interface for querying and controlling sensors."""

from decimal import Decimal

from .base import Interface


class TemperatureInterface(Interface):
    """Interface for querying and controlling sensors."""

    method_signatures = {
        "Temperature.GetValue": Decimal,
        "Temperature.GetValueHW": Decimal,
    }

    async def get_value(self, vid: int) -> Decimal:
        """Get the value of a temperature sensor, using cached value if available.

        Args:
            vid: The Vantage ID of the temperature sensor.

        Returns:
            The value of the temperature sensor, in degrees Celsius.
        """
        # INVOKE <id> Temperature.GetValue
        # -> R:INVOKE <id> <temp> Temperature.GetValue
        return await self.invoke(vid, "Temperature.GetValue", as_type=Decimal)

    async def get_value_hw(self, vid: int) -> Decimal:
        """Get the value of a temperature sensor.

        Args:
            vid: The Vantage ID of the temperature sensor.

        Returns:
            The value of the temperature sensor, in degrees Celsius.
        """
        # INVOKE <id> Temperature.GetValueHW
        # -> R:INVOKE <id> <temp> Temperature.GetValueHW
        return await self.invoke(vid, "Temperature.GetValueHW", as_type=Decimal)



"""IIntrospection.GetVersion method definition."""

from dataclasses import dataclass, field


@dataclass
class GetVersion:
    """IIntrospection.GetVersion method definition."""

    interface = "IIntrospection"

    @dataclass
    class Return:
        """Method return value."""

        kernel: str | None = field(default=None)
        rootfs: str | None = field(default=None)
        app: str | None = field(default=None)

    call = None
    return_value: Return | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""Controller holding and managing Vantage controllers."""

from contextlib import suppress
from typing import Any

from typing_extensions import override

from aiovantage.command_client.object_interfaces import (
    IntrospectionInterface,
    ObjectInterface,
)
from aiovantage.errors import CommandError
from aiovantage.models import Master

from .base import BaseController


class MastersController(
    BaseController[Master], IntrospectionInterface, ObjectInterface
):
    """Controller holding and managing Vantage controllers."""

    vantage_types = ("Master",)
    """The Vantage object types that this controller will fetch."""

    interface_status_types = ("Object.GetMTime",)
    """Which object interface status messages this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a Vantage controller."""
        state: dict[str, Any] = {}

        # IntrospectionInterface is not available on 2.x firmware.
        with suppress(CommandError):
            state[
                "firmware_version"
            ] = await IntrospectionInterface.get_firmware_version(
                self, vid, self.Firmware.Application
            )

        # ObjectInterface is not available on 2.x firmware.
        with suppress(CommandError):
            state["last_updated"] = await ObjectInterface.get_mtime(self, vid)

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        if method != "Object.GetMTime":
            return

        state = {
            "last_updated": self.parse_response(method, result, *args),
        }

        self.update_state(vid, state)



"""Command client object interfaces."""

from .anemo_sensor import AnemoSensorInterface
from .blind import BlindInterface
from .button import ButtonInterface
from .color_temperature import ColorTemperatureInterface
from .introspection import IntrospectionInterface
from .light_sensor import LightSensorInterface
from .load import LoadInterface
from .object import ObjectInterface
from .rgb_load import RGBLoadInterface
from .sensor import SensorInterface
from .sounder import SounderInterface
from .task import TaskInterface
from .temperature import TemperatureInterface
from .thermostat import ThermostatInterface

__all__ = [
    "AnemoSensorInterface",
    "BlindInterface",
    "ButtonInterface",
    "ColorTemperatureInterface",
    "IntrospectionInterface",
    "LightSensorInterface",
    "LoadInterface",
    "ObjectInterface",
    "RGBLoadInterface",
    "SensorInterface",
    "SounderInterface",
    "TaskInterface",
    "TemperatureInterface",
    "ThermostatInterface",
]



"""IIntrospection.GetSysInfo method definition."""

from dataclasses import dataclass, field


@dataclass
class GetSysInfo:
    """IIntrospection.GetSysInfo method definition."""

    interface = "IIntrospection"

    @dataclass
    class SysInfo:
        """SysInfo class."""

        master_number: int = field(metadata={"name": "MasterNumber"})
        serial_number: int = field(metadata={"name": "SerialNumber"})

    @dataclass
    class Return:
        """Method return value."""

        sys_info: "GetSysInfo.SysInfo" = field(metadata={"name": "SysInfo"})

    call = None
    return_value: Return | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""Base class for all station objects."""

from dataclasses import dataclass, field

from .location_object import LocationObject


@dataclass
class StationObject(LocationObject):
    """Base class for all station objects."""

    serial_number: str = field(
        metadata={
            "name": "SerialNumber",
        }
    )

    bus_id: int = field(
        metadata={
            "name": "Bus",
        }
    )



"""Interact with and control Vantage InFusion home automation controllers."""

__all__ = ["Vantage", "VantageEvent"]

import asyncio
from collections.abc import Callable
from types import TracebackType
from typing import Any, TypeVar, cast

from typing_extensions import Self

from .command_client import CommandClient, Event, EventStream, EventType
from .config_client import ConfigClient
from .controllers import (
    AnemoSensorsController,
    AreasController,
    BaseController,
    BlindGroupsController,
    BlindsController,
    ButtonsController,
    DryContactsController,
    GMemController,
    LightSensorsController,
    LoadGroupsController,
    LoadsController,
    MastersController,
    ModulesController,
    OmniSensorsController,
    PortDevicesController,
    PowerProfilesController,
    RGBLoadsController,
    StationsController,
    TasksController,
    TemperatureSensorsController,
    ThermostatsController,
)
from .events import EventCallback, VantageEvent
from .models import SystemObject

ControllerT = TypeVar("ControllerT", bound=BaseController[Any])


class Vantage:
    """Control a Vantage InFusion controller."""

    def __init__(
        self,
        host: str,
        username: str | None = None,
        password: str | None = None,
        *,
        use_ssl: bool = True,
        config_port: int | None = None,
        command_port: int | None = None,
    ) -> None:
        """Initialize the Vantage instance.

        Args:
            host: The hostname or IP address of the Vantage controller.
            username: The username to use for authentication.
            password: The password to use for authentication.
            use_ssl: Whether to use SSL for the connection.
            config_port: The port to use for the config client.
            command_port: The port to use for the command client.
        """
        # Set up clients
        self._host = host
        self._config_client = ConfigClient(
            host, username, password, ssl=use_ssl, port=config_port
        )

        self._command_client = CommandClient(
            host, username, password, ssl=use_ssl, port=command_port
        )

        self._event_stream = EventStream(
            host, username, password, ssl=use_ssl, port=command_port
        )

        # Set up controllers
        self._controllers: set[BaseController[Any]] = set()
        self._anemo_sensors = self._add_controller(AnemoSensorsController)
        self._areas = self._add_controller(AreasController)
        self._blind_groups = self._add_controller(BlindGroupsController)
        self._blinds = self._add_controller(BlindsController)
        self._buttons = self._add_controller(ButtonsController)
        self._dry_contacts = self._add_controller(DryContactsController)
        self._gmem = self._add_controller(GMemController)
        self._light_sensors = self._add_controller(LightSensorsController)
        self._load_groups = self._add_controller(LoadGroupsController)
        self._loads = self._add_controller(LoadsController)
        self._masters = self._add_controller(MastersController)
        self._modules = self._add_controller(ModulesController)
        self._rgb_loads = self._add_controller(RGBLoadsController)
        self._omni_sensors = self._add_controller(OmniSensorsController)
        self._port_devices = self._add_controller(PortDevicesController)
        self._power_profiles = self._add_controller(PowerProfilesController)
        self._stations = self._add_controller(StationsController)
        self._tasks = self._add_controller(TasksController)
        self._temperature_sensors = self._add_controller(TemperatureSensorsController)
        self._thermostats = self._add_controller(ThermostatsController)

        # Subscribe to reconnect events from the event stream
        self._event_stream.subscribe(self._handle_event, EventType.RECONNECTED)

    def __getitem__(self, vid: int) -> SystemObject:
        """Return the object with the given Vantage ID."""
        for controller in self._controllers:
            if vid in controller:
                return cast(SystemObject, controller[vid])
        raise KeyError(vid)

    def __contains__(self, vid: int) -> bool:
        """Is the given Vantage ID known by any controller."""
        return any(vid in controller for controller in self._controllers)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit context manager."""
        self.close()
        if exc_val:
            raise exc_val

    @property
    def host(self) -> str:
        """The hostname or IP address of the Vantage controller."""
        return self._host

    @property
    def config_client(self) -> ConfigClient:
        """The config client instance."""
        return self._config_client

    @property
    def command_client(self) -> CommandClient:
        """The command client instance."""
        return self._command_client

    @property
    def event_stream(self) -> EventStream:
        """The event stream instance."""
        return self._event_stream

    @property
    def anemo_sensors(self) -> AnemoSensorsController:
        """Controller for interacting with wind speed sensors."""
        return self._anemo_sensors

    @property
    def areas(self) -> AreasController:
        """Controller for interacting with areas."""
        return self._areas

    @property
    def blinds(self) -> BlindsController:
        """Controller for interacting with blinds."""
        return self._blinds

    @property
    def blind_groups(self) -> BlindGroupsController:
        """Controller for interacting with groups of blinds."""
        return self._blind_groups

    @property
    def buttons(self) -> ButtonsController:
        """Controller for interacting with keypad buttons."""
        return self._buttons

    @property
    def dry_contacts(self) -> DryContactsController:
        """Controller for interacting with dry contacts."""
        return self._dry_contacts

    @property
    def gmem(self) -> GMemController:
        """Controller for interacting with variables."""
        return self._gmem

    @property
    def light_sensors(self) -> LightSensorsController:
        """Controller for interacting with light sensors."""
        return self._light_sensors

    @property
    def loads(self) -> LoadsController:
        """Controller for interacting with loads (lights, fans, etc)."""
        return self._loads

    @property
    def load_groups(self) -> LoadGroupsController:
        """Controller for interacting with groups of loads."""
        return self._load_groups

    @property
    def masters(self) -> MastersController:
        """Controller for interacting with Vantage Controllers."""
        return self._masters

    @property
    def modules(self) -> ModulesController:
        """Controller for interacting with dimmer modules."""
        return self._modules

    @property
    def omni_sensors(self) -> OmniSensorsController:
        """Controller for interacting with "omni" sensors."""
        return self._omni_sensors

    @property
    def port_devices(self) -> PortDevicesController:
        """Controller for interacting with port devices."""
        return self._port_devices

    @property
    def power_profiles(self) -> PowerProfilesController:
        """Controller for interacting with power profiles."""
        return self._power_profiles

    @property
    def rgb_loads(self) -> RGBLoadsController:
        """Controller for interacting with RGB loads."""
        return self._rgb_loads

    @property
    def stations(self) -> StationsController:
        """Controller for interacting with stations (keypads, etc)."""
        return self._stations

    @property
    def tasks(self) -> TasksController:
        """Controller for interacting with tasks."""
        return self._tasks

    @property
    def temperature_sensors(self) -> TemperatureSensorsController:
        """Controller for interacting with temperature sensors."""
        return self._temperature_sensors

    @property
    def thermostats(self) -> ThermostatsController:
        """Controller for interacting with thermostats."""
        return self._thermostats

    def get(self, vid: int) -> SystemObject | None:
        """Return the object with the given Vantage ID, or None if not found."""
        try:
            return self[vid]
        except KeyError:
            return None

    def close(self) -> None:
        """Close the clients."""
        self.config_client.close()
        self.command_client.close()
        self.event_stream.stop()

    async def initialize(self, fetch_state: bool = True) -> None:
        """Fetch all objects from the controllers.

        Args:
            fetch_state: Whether to also fetch the state of each object.
        """
        # Initialize all controllers
        await asyncio.gather(
            *[controller.initialize(fetch_state) for controller in self._controllers]
        )

        # Start the event stream
        await self.event_stream.start()

    def subscribe(self, callback: EventCallback[SystemObject]) -> Callable[[], None]:
        """Subscribe to state changes for all objects.

        Args:
            callback: The callback to call when an object changes.

        Returns:
            A function to unsubscribe.
        """
        unsubscribes = [
            controller.subscribe(callback) for controller in self._controllers
        ]

        def unsubscribe() -> None:
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def _handle_event(self, event: Event) -> None:
        # Handle events from the event stream.
        if event["type"] == EventType.RECONNECTED:
            for controller in self._controllers:
                if controller.initialized:
                    await controller.fetch_full_state()

    def _add_controller(self, controller_cls: type[ControllerT]) -> ControllerT:
        # Add a controller to the known controllers.
        controller = controller_cls(self)
        self._controllers.add(controller)
        return controller



"""Temperature object."""

from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum

from .sensor import Sensor
from .types import Parent


@dataclass(kw_only=True)
class Temperature(Sensor):
    """Temperature object."""

    class Setpoint(Enum):
        """Setpoint type."""

        HEAT = "Heat"
        COOL = "Cool"
        AUTO = "Auto"

    # Not available in 2.x firmware
    setpoint: Setpoint | None = field(
        default=None,
        metadata={
            "name": "Setpoint",
            "type": "Attribute",
        },
    )

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    value: Decimal | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Controller holding and managing Vantage blinds."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.object_interfaces import BlindInterface
from aiovantage.models import BlindBase

from .base import BaseController


class BlindsController(BaseController[BlindBase], BlindInterface):
    """Controller holding and managing Vantage blinds."""

    vantage_types = (
        "QISBlind",
        "QubeBlind",
        "RelayBlind",
        "Somfy.RS-485_Shade_CHILD",
        "Somfy.URTSI_2_Shade_CHILD",
    )
    """The Vantage object types that this controller will fetch."""

    status_types = ("BLIND",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a blind."""
        state = {
            "position": await BlindInterface.get_position(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "BLIND":
            return

        # STATUS BLIND
        # -> S:BLIND <id> <position (0.000 - 100.000)>
        state = {
            "position": Decimal(args[0]),
        }

        self.update_state(vid, state)



"""IIntrospection.GetTypes method definition."""

from dataclasses import dataclass, field


@dataclass
class GetTypes:
    """IIntrospection.GetTypes method definition."""

    interface = "IIntrospection"

    @dataclass
    class Type:
        """Object type definition."""

        name: str = field(metadata={"name": "Name"})
        version: str = field(metadata={"name": "Version"})

    call = None
    return_value: list[Type] | None = field(
        default=None,
        metadata={
            "name": "Type",
            "wrapper": "return",
        },
    )



"""DMX Gateway color load object."""

from dataclasses import dataclass

from .rgb_load_base import RGBLoadBase


@dataclass
class VantageDGColorLoad(RGBLoadBase):
    """DMX Gateway color load object."""

    class Meta:
        name = "Vantage.DGColorLoad"



"""Client for communicating with the Vantage Host Command service.

The Host Command service is a text-based service that allows interaction with devices
controlled by a Vantage InFusion Controller.

Among other things, this service allows you to change the state of devices
(eg. turn on/off a light) as well as subscribe to status changes for devices.

The service is exposed on port 3010 (SSL) by default, and on port 3001 (non-SSL) if this
port has been opened by the firewall on the controller.

The service is discoverable via mDNS as `_hc._tcp.local` and/or `_secure_hc._tcp.local`.
"""

from .commands import CommandClient
from .events import Event, EventStream, EventType

__all__ = [
    "CommandClient",
    "Event",
    "EventStream",
    "EventType",
]



"""Interface for querying and controlling loads."""

from decimal import Decimal
from enum import IntEnum

from .base import Interface


class LoadInterface(Interface):
    """Interface for querying and controlling loads."""

    class RampType(IntEnum):
        """The type of ramp to perform."""

        Up = 5
        Down = 4
        Opposite = 3
        Stop = 2
        Fixed = 6
        Variable = 7
        Adjust = 8

    method_signatures = {
        "Load.GetLevel": Decimal,
        "Load.GetLevelHW": Decimal,
    }

    async def set_level(self, vid: int, level: float | Decimal) -> None:
        """Set the level of a load.

        Args:
            vid: The Vantage ID of the load.
            level: The level to set the load to (0-100).
        """
        # Clamp level to 0-100
        level = max(min(level, 100), 0)

        # INVOKE <id> Load.SetLevel <level (0-100)>
        # -> R:INVOKE <id> <rcode> Load.SetLevel <level (0-100)>
        await self.invoke(vid, "Load.SetLevel", level)

    async def get_level(self, vid: int) -> Decimal:
        """Get the level of a load, using cached value if available.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetLevel
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetLevel
        return await self.invoke(vid, "Load.GetLevel", as_type=Decimal)

    async def get_level_hw(self, vid: int) -> Decimal:
        """Get the level of a load directly from the hardware.

        Args:
            vid: The Vantage ID of the load.

        Returns:
            The level of the load, as a percentage (0-100).
        """
        # INVOKE <id> Load.GetLevelHW
        # -> R:INVOKE <id> <level (0.000-100.000)> Load.GetLevelHW
        return await self.invoke(vid, "Load.GetLevelHW", as_type=Decimal)

    async def ramp(
        self,
        vid: int,
        cmd: RampType = RampType.Fixed,
        time: float | Decimal = 0,
        level: float | Decimal = 0,
    ) -> None:
        """Ramp a load to a level over a number of seconds.

        Args:
            vid: The Vantage ID of the load.
            cmd: The type of ramp to perform.
            time: The number of seconds to ramp the load over.
            level: The level to ramp the load to (0-100).
        """
        # Clamp level to 0-100
        level = max(min(level, 100), 0)

        # INVOKE <id> Load.Ramp <cmd> <time> <level>
        # -> R:INVOKE <id> <rcode> Load.Ramp <cmd> <time> <level>
        await self.invoke(vid, "Load.Ramp", cmd, time, level)

    # Additional convenience methods, not part of the Vantage API
    async def turn_on(
        self, vid: int, transition: float | None = None, level: float | None = None
    ) -> None:
        """Turn on a load with an optional transition time.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to transition to the new level, defaults to immediate.
            level: The level to set the load to (0-100), defaults to 100.
        """
        if level is None:
            level = 100

        if transition is None:
            return await self.set_level(vid, level)

        await self.ramp(vid, time=transition, level=level)

    async def turn_off(self, vid: int, transition: float | None = None) -> None:
        """Turn off a load with an optional transition time.

        Args:
            vid: The Vantage ID of the load.
            transition: The time in seconds to ramp the load down, defaults to immediate.
        """
        if transition is None:
            return await self.set_level(vid, 0)

        await self.ramp(vid, time=transition, level=0)



"""Somfy RS-485 SDN 2.0."""
from dataclasses import dataclass

from .port_device import PortDevice


@dataclass
class SomfyRS485SDN20Port(PortDevice):
    """Somfy RS-485 SDN 2.0."""

    class Meta:
        name = "Somfy.RS-485_SDN_2_x2E_0_PORT"



"""Power profile object."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass
class PowerProfile(SystemObject):
    """Power Profile object."""

    min: float = field(
        metadata={
            "name": "Min",
        }
    )

    max: float = field(
        metadata={
            "name": "Max",
        }
    )

    # Not available in 2.x firmware
    adjust: int | None = field(
        default=None,
        metadata={
            "name": "Adjust",
        },
    )

    # Not available in 2.x firmware
    freq: int | None = field(
        default=None,
        metadata={
            "name": "Freq",
        },
    )

    # Not available in 2.x firmware
    inductive: bool | None = field(
        default=None,
        metadata={
            "name": "Inductive",
        },
    )

    @property
    def is_dimmable(self) -> bool:
        """Return True if loads with this profile are dimmable."""
        return self.max > self.min



"""Controller holding and managing Vantage omni sensors."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.utils import parse_fixed_param
from aiovantage.config_client.models.omni_sensor import ConversionType, OmniSensor

from .base import BaseController


class OmniSensorsController(BaseController[OmniSensor]):
    """Controller holding and managing Vantage omni sensors.

    Omni sensors are generic sensors objects which specify which methods to use
    when getting or setting data in their object definition, as well as the
    type of data and a conversion formula.
    """

    vantage_types = ("OmniSensor",)
    """The Vantage object types that this controller will fetch."""

    interface_status_types = "*"
    """Which object interface status messages this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of an omni sensor."""
        state = {
            "level": await self.get_level(vid),
        }

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *_args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        omni_sensor = self[vid]
        if method != omni_sensor.get.method:
            return

        state = {
            "level": self.parse_result(omni_sensor, result),
        }

        self.update_state(vid, state)

    async def get_level(self, vid: int, cached: bool = True) -> int | Decimal:
        """Get the level of an OmniSensor.

        Args:
            vid: The ID of the sensor.
            cached: Whether to use the cached value or fetch a new one.

        Returns:
            The level of the sensor.
        """
        omni_sensor = self[vid]

        # INVOKE <id> <method>
        # -> R:INVOKE <id> <value> <method>
        method = omni_sensor.get.method if cached else omni_sensor.get.method_hw
        response = await self.command_client.command("INVOKE", vid, method)

        return self.parse_result(omni_sensor, response.args[1])

    @classmethod
    def parse_result(cls, sensor: OmniSensor, result: str) -> int | Decimal:
        """Parse an OmniSensor response, eg. 'PowerSensor.GetPower'."""
        # NOTE: This currently doesn't handle conversion formulas, or return_type
        level = parse_fixed_param(result)
        if sensor.get.formula.level_type == ConversionType.INT:
            return int(level)

        return level



"""Equinox 41 or Equinox 73 touchscreen."""

from dataclasses import dataclass

from .station_object import StationObject


@dataclass
class EqUX(StationObject):
    """Equinox 41 or Equinox 73 touchscreen."""



"""Controller holding and managing Vantage variables."""
import re

from typing_extensions import override

from aiovantage.command_client.utils import parse_byte_param
from aiovantage.models import GMem

from .base import BaseController


class GMemController(BaseController[GMem]):
    """Controller holding and managing Vantage variables.

    We use the `GETVARIABLE` and `VARIABLE` wrappers for getting and setting
    variable values, rather than the GMem object interface, since they are much
    simpler than working with raw byte arrays.
    """

    vantage_types = ("GMem",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("VARIABLE",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a variable."""
        state = {
            "value": await self.get_value(vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "VARIABLE":
            return

        # STATUS VARIABLE
        # -> S:VARIABLE <id> <value>
        state = {
            "value": self._parse_value(vid, args[0]),
        }

        self.update_state(vid, state)

    async def get_value(self, vid: int) -> int | str | bool:
        """Get the value of a variable.

        Args:
            vid: The Vantage ID of the variable.

        Returns:
            The value of the variable, either a bool, int, or str.
        """
        # GETVARIABLE {id}
        # -> R:GETVARIABLE {id} {value}
        response = await self.command_client.command("GETVARIABLE", vid)
        raw_value = response.args[1]

        return self._parse_value(vid, raw_value)

    async def set_value(self, vid: int, value: int | str | bool) -> None:
        """Set the value of a variable.

        Args:
            vid: The Vantage ID of the variable.
            value: The value to set, either a bool, int, or str.
        """
        # VARIABLE {id} {value}
        # -> R:VARIABLE {id} {value}
        await self.command_client.command("VARIABLE", vid, value, force_quotes=True)

    def _parse_value(self, vid: int, value: str) -> int | str | bool:
        # Parse the results of a GMem lookup into the expected type.
        gmem: GMem = self[vid]

        if gmem.is_bool:
            return bool(int(value))
        if gmem.is_str:
            # Handle byte array strings
            if re.match(r"^[\[\{].*[\]\}]$", value):
                byte_param = parse_byte_param(value)
                return byte_param.decode().rstrip("\x00")

            return value

        return int(value)



"""Master (controller) object."""

from dataclasses import dataclass, field

from .system_object import SystemObject


@dataclass
class Master(SystemObject):
    """Master (controller) object."""

    number: int = field(
        metadata={
            "name": "Number",
        }
    )

    volts: float = field(
        metadata={
            "name": "Volts",
        }
    )

    amps: float = field(
        metadata={
            "name": "Amps",
        }
    )

    # Not available in 2.x firmware
    module_count: int | None = field(
        default=None,
        metadata={
            "name": "ModuleCount",
        },
    )

    # Not available in 2.x firmware
    serial_number: int | None = field(
        default=None,
        metadata={
            "name": "SerialNumber",
        },
    )

    firmware_version: str | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    last_updated: int | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Vantage object types."""

from .anemo_sensor import AnemoSensor
from .area import Area
from .blind import Blind
from .blind_base import BlindBase
from .blind_group import BlindGroup
from .blind_group_base import BlindGroupBase
from .button import Button
from .child_device import ChildDevice
from .custom_device import CustomDevice
from .dc_power_profile import DCPowerProfile
from .dimmer import Dimmer
from .dry_contact import DryContact
from .dual_relay_station import DualRelayStation
from .eq_ctrl import EqCtrl
from .eq_ux import EqUX
from .gmem import GMem
from .keypad import Keypad
from .light_sensor import LightSensor
from .load import Load
from .load_group import LoadGroup
from .location_object import LocationObject
from .master import Master
from .module import Module
from .module_gen2 import ModuleGen2
from .omni_sensor import OmniSensor
from .parent_device import ParentDevice
from .port_device import PortDevice
from .power_profile import PowerProfile
from .pwm_power_profile import PWMPowerProfile
from .qis_blind import QISBlind
from .qube_blind import QubeBlind
from .relay_blind import RelayBlind
from .rgb_load_base import RGBLoadBase
from .scene_point_relay import ScenePointRelay
from .sensor import Sensor
from .somfy_rs_485_group_child import SomfyRS485GroupChild
from .somfy_rs_485_sdn_20_port import SomfyRS485SDN20Port
from .somfy_rs_485_shade_child import SomfyRS485ShadeChild
from .somfy_urtsi_2_group_child import SomfyURTSI2GroupChild
from .somfy_urtsi_2_port import SomfyURTSI2Port
from .somfy_urtsi_2_shade_child import SomfyURTSI2ShadeChild
from .station_bus import StationBus
from .station_object import StationObject
from .system_object import SystemObject
from .task import Task
from .temperature import Temperature
from .thermostat import Thermostat
from .types import Parent
from .vantage_ddg_color_load import VantageDDGColorLoad
from .vantage_dg_color_load import VantageDGColorLoad
from .vantage_dmx_dali_gateway import VantageDmxDaliGateway
from .vantage_dmx_gateway import VantageDmxGateway

__all__ = [
    "AnemoSensor",
    "Area",
    "Blind",
    "BlindBase",
    "BlindGroup",
    "BlindGroupBase",
    "Button",
    "ChildDevice",
    "CustomDevice",
    "DCPowerProfile",
    "Dimmer",
    "DryContact",
    "DualRelayStation",
    "EqCtrl",
    "EqUX",
    "GMem",
    "Keypad",
    "LightSensor",
    "Load",
    "LoadGroup",
    "LocationObject",
    "Master",
    "Module",
    "ModuleGen2",
    "OmniSensor",
    "Parent",
    "ParentDevice",
    "PortDevice",
    "PowerProfile",
    "PWMPowerProfile",
    "QISBlind",
    "QubeBlind",
    "RelayBlind",
    "RGBLoadBase",
    "ScenePointRelay",
    "Sensor",
    "SomfyRS485GroupChild",
    "SomfyRS485SDN20Port",
    "SomfyRS485ShadeChild",
    "SomfyURTSI2GroupChild",
    "SomfyURTSI2Port",
    "SomfyURTSI2ShadeChild",
    "StationBus",
    "StationObject",
    "SystemObject",
    "Task",
    "Temperature",
    "Thermostat",
    "VantageDDGColorLoad",
    "VantageDGColorLoad",
    "VantageDmxDaliGateway",
    "VantageDmxGateway",
]



"""Controller holding and managing Vantage temperature sensors."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.object_interfaces import TemperatureInterface
from aiovantage.models import Temperature

from .base import BaseController


class TemperatureSensorsController(BaseController[Temperature], TemperatureInterface):
    """Controller holding and managing Vantage temperature sensors."""

    vantage_types = ("Temperature",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("TEMP",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a temperature sensor."""
        state = {
            "value": await TemperatureInterface.get_value(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status message from the event stream."""
        if status != "TEMP":
            return

        # STATUS TEMP
        # -> S:TEMP <id> <temp>
        state = {
            "value": Decimal(args[0]),
        }

        self.update_state(vid, state)



"""Module object."""

from dataclasses import dataclass, field

from .system_object import SystemObject
from .types import Parent


@dataclass
class Module(SystemObject):
    """Module object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )



"""Duck typing for methods."""

from typing import Protocol, TypeVar

Call = TypeVar("Call")
Return = TypeVar("Return")


class Method(Protocol[Call, Return]):
    """Duck typing for methods."""

    interface: str
    call: Call | None
    return_value: Return | None



"""Dump all Vantage object types."""
import argparse
import asyncio
import contextlib
from pathlib import Path
from xml.etree import ElementTree

from aiovantage.config_client import ConfigClient
from aiovantage.config_client.interfaces.introspection import GetInterfaces, GetTypes
from aiovantage.discovery import get_controller_details

# Grab connection info from command line arguments
parser = argparse.ArgumentParser()
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
args = parser.parse_args()

ElementTree.register_namespace("vntg", "http://www.vantagecontrols.com/Vantage")
ElementTree.register_namespace("xs", "http://www.w3.org/2001/XMLSchema")


async def get_type_info(client: ConfigClient, object_type: str) -> str:
    """Get the XML schema for the specified object types."""
    response = await client.raw_request(
        "IIntrospection",
        f"<GetTypeInfo><call><Type>{object_type}</Type></call></GetTypeInfo>",
    )

    root = ElementTree.fromstring(
        response,
        parser=ElementTree.XMLParser(
            target=ElementTree.TreeBuilder(insert_pis=True, insert_comments=True)
        ),
    )
    return_el = root.find("./GetTypeInfo/return/TypeInfo")
    assert return_el is not None

    doc = '<?xml version="1.0" encoding="UTF-8"?>\n'
    for child in return_el:
        doc += ElementTree.tostring(child, encoding="unicode")
    return doc


async def get_interface_info(client: ConfigClient, interface: str) -> str:
    """Get the XML schema for the specified object types."""
    response = await client.raw_request(
        "IIntrospection",
        f"<GetInterfaceInfo><call><Interface>{interface}</Interface></call></GetInterfaceInfo>",
    )

    root = ElementTree.fromstring(
        response,
        parser=ElementTree.XMLParser(
            target=ElementTree.TreeBuilder(insert_pis=True, insert_comments=True)
        ),
    )
    return_el = root.find("./GetInterfaceInfo/return/InterfaceInfo")
    assert return_el is not None

    doc = '<?xml version="1.0" encoding="UTF-8"?>\n'
    for child in return_el:
        doc += ElementTree.tostring(child, encoding="unicode")
    return doc


async def main() -> None:
    """Dump all Vantage object types."""
    controller = await get_controller_details(args.host)
    if controller is None:
        parser.print_usage()
        print(f"error: no Vantage controller found at {args.host}")
        return

    if controller.requires_auth and not (args.username or args.password):
        parser.print_usage()
        print("error: controller requires auth, but no username/password provided")
        return

    async with ConfigClient(
        args.host, args.username, args.password, ssl=controller.supports_ssl
    ) as client:
        # Dump all object type schemas
        types = await client.request(GetTypes)
        if types is None:
            print("error: IIntrospection.GetTypes returned None")
            return

        for object_type in types:
            try:
                schema = await get_type_info(client, object_type.name)
                file = Path(f"schemas/types/{object_type.name}.xsd")
                file.parent.mkdir(parents=True, exist_ok=True)
                file.write_text(schema, encoding="utf-8")
            except ElementTree.ParseError as exc:
                print(f"Failed to get type info for {object_type.name}: {exc}")

        # Dump all object interface definitions
        interfaces = await client.request(GetInterfaces)
        if interfaces is None:
            print("error: IIntrospection.GetInterfaces returned None")
            return

        for interface in interfaces:
            try:
                schema = await get_interface_info(client, interface.name)
                file = Path(f"schemas/interfaces/{interface.name}.vidl")
                file.parent.mkdir(parents=True, exist_ok=True)
                file.write_text(schema, encoding="utf-8")
            except ElementTree.ParseError as exc:
                print(f"Failed to get interface info for {interface.name}: {exc}")

        print(f"Dumped types and interfaces to {Path.cwd() / 'schemas'}")


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""IConfiguration.GetObject method definition."""

from dataclasses import dataclass, field

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetObject:
    """IConfiguration.GetObject method definition."""

    interface = "IConfiguration"

    @dataclass
    class Params:
        """Method parameters."""

        vids: list[int] = field(
            metadata={
                "name": "VID",
            }
        )

    call: Params | None = field(default=None)
    return_value: list[ObjectChoice] | None = field(
        default=None,
        metadata={
            "name": "Object",
            "wrapper": "return",
        },
    )



"""DryContact object."""

from dataclasses import dataclass, field

from .location_object import LocationObject
from .types import Parent


@dataclass
class DryContact(LocationObject):
    """DryContact object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    triggered: bool | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Blind object."""

from dataclasses import dataclass, field

from .blind_base import BlindBase
from .location_object import LocationObject
from .types import Parent


@dataclass
class Blind(BlindBase, LocationObject):
    """Blind object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )



"""Thermostat object."""
from dataclasses import dataclass, field
from enum import IntEnum

from .station_object import StationObject


@dataclass
class Thermostat(StationObject):
    """Thermostat object."""

    class OperationMode(IntEnum):
        """The operation mode of the thermostat."""

        OFF = 0
        COOL = 1
        HEAT = 2
        AUTO = 3

    class FanMode(IntEnum):
        """The fan mode of the thermostat."""

        AUTO = 0
        ON = 1

    class DayMode(IntEnum):
        """The day mode of the thermostat."""

        DAY = 0
        NIGHT = 1

    class HoldMode(IntEnum):
        """The hold mode of the thermostat."""

        NORMAL = 0
        HOLD = 1

    class Status(IntEnum):
        """The status of the thermostat."""

        OFF = 0
        COOLING = 1
        HEATING = 2
        OFFLINE = 3

    operation_mode: OperationMode | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    fan_mode: FanMode | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    day_mode: DayMode | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    hold_mode: HoldMode | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    # Not available in 2.x firmware
    status: Status | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Controller holding and managing thermostats."""

from contextlib import suppress
from typing import Any

from typing_extensions import override

from aiovantage.command_client.object_interfaces import ThermostatInterface
from aiovantage.errors import CommandError
from aiovantage.models import Temperature, Thermostat
from aiovantage.query import QuerySet

from .base import BaseController


class ThermostatsController(BaseController[Thermostat], ThermostatInterface):
    """Controller holding and managing thermostats.

    Thermostats have a number of temperature sensors associated with them which
    represent the current indoor temperature, outdoor temperature, and the
    current cool and heat setpoints.
    """

    vantage_types = ("Thermostat",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("THERMFAN", "THERMOP", "THERMDAY")
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a thermostat."""
        state = {
            "operation_mode": await ThermostatInterface.get_operation_mode(self, vid),
            "fan_mode": await ThermostatInterface.get_fan_mode(self, vid),
            "day_mode": await ThermostatInterface.get_day_mode(self, vid),
        }

        # Hold mode is not supported by every thermostat type.
        with suppress(CommandError):
            state["hold_mode"] = await ThermostatInterface.get_hold_mode(self, vid)

        # Status is not available on 2.x firmware.
        with suppress(CommandError):
            state["status"] = await ThermostatInterface.get_status(self, vid)

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status message from the event stream."""
        state: dict[str, Any] = {}

        if status == "THERMOP":
            # STATUS THERMOP
            # -> S:THERMOP <id> <operation_mode (OFF/COOL/HEAT/AUTO)>
            state["operation_mode"] = Thermostat.OperationMode[args[0]]
        elif status == "THERMFAN":
            # STATUS THERMFAN
            # -> S:THERMFAN <id> <fan_mode (ON/AUTO)>
            state["fan_mode"] = Thermostat.FanMode[args[0]]
        elif status == "THERMDAY":
            # STATUS THERMDAY
            # -> S:THERMDAY <id> <day_mode (DAY/NIGHT)>
            state["day_mode"] = Thermostat.DayMode[args[0]]
        else:
            return

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        state: dict[str, Any] = {}
        if method == "Thermostat.GetHoldMode":
            state["hold_mode"] = self.parse_response(method, result, *args)
        elif method == "Thermostat.GetStatus":
            state["status"] = self.parse_response(method, result, *args)

        self.update_state(vid, state)

    def sensors(self, vid: int) -> QuerySet[Temperature]:
        """Return all sensors associated with this thermostat."""
        return self._vantage.temperature_sensors.filter(
            lambda obj: obj.parent.id == vid
        )

    def indoor_sensor(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the indoor temperature sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 1)

    def outdoor_sensor(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the outdoor temperature sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 2)

    def cool_setpoint(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the cool setpoint sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 3)

    def heat_setpoint(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the heat setpoint sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 4)



"""Button object."""

from dataclasses import dataclass, field

from .system_object import SystemObject
from .types import Parent


@dataclass
class Button(SystemObject):
    """Button object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    down_id: int = field(
        metadata={
            "name": "Down",
        }
    )

    up_id: int = field(
        metadata={
            "name": "Up",
        }
    )

    hold_id: int = field(
        metadata={
            "name": "Hold",
        }
    )

    text1: str = field(
        metadata={
            "name": "Text1",
        }
    )

    text2: str = field(
        metadata={
            "name": "Text2",
        }
    )

    pressed: bool | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )

    @property
    def text(self) -> str:
        """Return the button text."""
        return f"{self.text1}\n{self.text2}" if self.text2 else self.text1



"""ILogin.Login method definition."""

from dataclasses import dataclass, field


@dataclass
class Login:
    """ILogin.Login method definition."""

    interface = "ILogin"

    @dataclass
    class Params:
        """Method parameters."""

        user: str = field(
            metadata={
                "name": "User",
            }
        )

        password: str = field(
            metadata={
                "name": "Password",
            }
        )

    call: Params | None = field(default=None)
    return_value: bool | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""Sensor object."""

from dataclasses import dataclass

from .location_object import LocationObject

# NOTE: Firmware 2.x inherits from SystemObject, not LocationObject


@dataclass
class Sensor(LocationObject):
    """Sensor object."""



"""Controller holding and managing Vantage blind groups."""

from aiovantage.command_client.object_interfaces import BlindInterface
from aiovantage.models import BlindBase, BlindGroup, BlindGroupBase, ChildDevice
from aiovantage.query import QuerySet

from .base import BaseController


class BlindGroupsController(BaseController[BlindGroupBase], BlindInterface):
    """Controller holding and managing Vantage blind groups."""

    vantage_types = (
        "BlindGroup",
        "Somfy.RS-485_Group_CHILD",
        "Somfy.URTSI_2_Group_CHILD",
    )
    """The Vantage object types that this controller will fetch."""

    def blinds(self, vid: int) -> QuerySet[BlindBase]:
        """Return a queryset of all blinds in this blind group."""
        blind_group = self[vid]
        if isinstance(blind_group, BlindGroup):

            def _filter1(blind: BlindBase) -> bool:
                return blind.id in blind_group.blind_ids

            return self._vantage.blinds.filter(_filter1)

        # Otherwise, use the parent id to filter
        def _filter2(blind: BlindBase) -> bool:
            if isinstance(blind, ChildDevice):
                return blind.parent.id == blind_group.id

            return False

        return self._vantage.blinds.filter(_filter2)



"""Somfy URTSI 2 blind."""

from dataclasses import dataclass

from .blind_base import BlindBase
from .child_device import ChildDevice


@dataclass
class SomfyURTSI2ShadeChild(BlindBase, ChildDevice):
    """Somfy URTSI 2 blind."""

    class Meta:
        name = "Somfy.URTSI_2_Shade_CHILD"



"""Common types."""

from dataclasses import dataclass, field


@dataclass
class Parent:
    """Parent tag."""

    id: int

    position: int = field(
        metadata={
            "name": "Position",
            "type": "Attribute",
        }
    )



"""ObjectChoice type definition."""

import inspect
from dataclasses import dataclass, field
from functools import cache
from types import ModuleType
from typing import Any

from aiovantage.config_client import models


def get_all_module_classes(module: ModuleType) -> list[type[Any]]:
    """Get all classes from a module."""
    classes = []
    for _, cls in inspect.getmembers(module, inspect.isclass):
        classes.append(cls)

    return classes


@cache
def get_all_object_choices(module: ModuleType) -> list[dict[str, Any]]:
    """Get all object choices from a module."""
    choices = []
    for cls in get_all_module_classes(module):
        # Get the name of the XML element
        meta = getattr(cls, "Meta", None)
        name = getattr(meta, "name", cls.__qualname__)

        # Add the xsdata choice
        choices.append({"name": name, "type": cls})

    return choices


@dataclass
class ObjectChoice:
    """Wildcard type that can be used to represent any object."""

    id: int = field(
        metadata={
            "name": "VID",
            "type": "Attribute",
        }
    )

    choice: object = field(
        metadata={
            "type": "Wildcard",
            "choices": get_all_object_choices(models),
        },
    )



"""Controller holding and managing Vantage RGB loads."""

from typing import Any

from typing_extensions import override

from aiovantage.command_client.object_interfaces import (
    ColorTemperatureInterface,
    LoadInterface,
    RGBLoadInterface,
)
from aiovantage.models import RGBLoadBase
from aiovantage.query import QuerySet

from .base import BaseController


class RGBLoadsController(
    BaseController[RGBLoadBase],
    LoadInterface,
    RGBLoadInterface,
    ColorTemperatureInterface,
):
    """Controller holding and managing Vantage RGB loads."""

    vantage_types = ("Vantage.DGColorLoad", "Vantage.DDGColorLoad")
    """The Vantage object types that this controller will fetch."""

    interface_status_types = (
        "Load.GetLevel",
        "RGBLoad.GetHSL",
        "RGBLoad.GetRGB",
        "RGBLoad.GetRGBW",
        "ColorTemperature.Get",
    )
    """Which object interface status messages this controller handles, if any."""

    def __post_init__(self) -> None:
        """Initialize the map for building colors."""
        self._temp_color_map: dict[int, list[int]] = {}

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of an RGB load."""
        state: dict[str, Any] = {
            "level": await LoadInterface.get_level(self, vid),
        }

        rgb_load: RGBLoadBase = self[vid]
        if rgb_load.is_rgb:
            state["hsl"] = await RGBLoadInterface.get_hsl_color(self, vid)
            state["rgb"] = await RGBLoadInterface.get_rgb_color(self, vid)
            state["rgbw"] = await RGBLoadInterface.get_rgbw_color(self, vid)

        if rgb_load.is_cct:
            state["color_temp"] = await ColorTemperatureInterface.get_color_temp(
                self, vid
            )

        self.update_state(vid, state)

    @override
    def handle_interface_status(
        self, vid: int, method: str, result: str, *args: str
    ) -> None:
        """Handle object interface status messages from the event stream."""
        rgb_load: RGBLoadBase = self[vid]
        state: dict[str, Any] = {}

        if method == "Load.GetLevel":
            state["level"] = self.parse_response(method, result, *args)

        elif method == "RGBLoad.GetHSL" and rgb_load.is_rgb:
            if color := self._parse_color_channel_response(vid, method, result, *args):
                state["hsl"] = color

        elif method == "RGBLoad.GetRGB" and rgb_load.is_rgb:
            if color := self._parse_color_channel_response(vid, method, result, *args):
                state["rgb"] = color

        elif method == "RGBLoad.GetRGBW" and rgb_load.is_rgb:
            if color := self._parse_color_channel_response(vid, method, result, *args):
                state["rgbw"] = color

        elif method == "ColorTemperature.Get" and rgb_load.is_cct:
            state["color_temp"] = self.parse_response(method, result, *args)

        else:
            return

        self.update_state(vid, state)

    @property
    def is_on(self) -> QuerySet[RGBLoadBase]:
        """Return a queryset of all RGB loads that are turned on."""
        return self.filter(lambda load: load.is_on)

    @property
    def is_off(self) -> QuerySet[RGBLoadBase]:
        """Return a queryset of all RGB loads that are turned off."""
        return self.filter(lambda load: not load.is_on)

    def _parse_color_channel_response(
        self, vid: int, method: str, result: str, *args: str
    ) -> tuple[int, ...] | None:
        # Build a color from a series of channel responses. We need to store
        # partially constructed colors in memory, since updates come separately for
        # each channel.

        # Determine how many channels we need to build a color
        if method in ("RGBLoad.GetHSL", "RGBLoad.GetRGB"):
            num_channels = 3
        elif method == "RGBLoad.GetRGBW":
            num_channels = 4
        else:
            raise ValueError(f"Unsupported color channel method {method}")

        # Parse the response
        response = self.parse_response(
            method, result, *args, as_type=self.ColorChannelResponse
        )

        # Ignore updates for channels we don't care about
        if response.channel < 0 or response.channel >= num_channels:
            return None

        # Store the channel value in the temp color map
        self._temp_color_map.setdefault(vid, num_channels * [0])
        self._temp_color_map[vid][response.channel] = response.value

        # If we have all the channels, build and return the color
        if response.channel == num_channels - 1:
            color = tuple(self._temp_color_map[vid])
            del self._temp_color_map[vid]
            return color

        return None



"""Base class for parent device objects."""

from dataclasses import dataclass

from .custom_device import CustomDevice


@dataclass
class ParentDevice(CustomDevice):
    """Base class for parent device objects."""



"""Interface for querying and controlling light sensors."""

from decimal import Decimal

from .base import Interface


class LightSensorInterface(Interface):
    """Interface for querying and controlling light sensors."""

    method_signatures = {
        "LightSensor.GetLevel": Decimal,
        "LightSensor.GetLevelHW": Decimal,
    }

    async def get_level(self, vid: int) -> Decimal:
        """Get the level of a light sensor, using cached value if available.

        Args:
            vid: The Vantage ID of the light sensor.

        Returns:
            The level of the light sensor, in foot-candles.
        """
        # INVOKE <id> LightSensor.GetLevel
        # -> R:INVOKE <id> <level> LightSensor.GetLevel
        return await self.invoke(vid, "LightSensor.GetLevel", as_type=Decimal)

    async def get_level_hw(self, vid: int) -> Decimal:
        """Get the level of a light sensor directly from the hardware.

        Args:
            vid: The Vantage ID of the light sensor.

        Returns:
            The level of the light sensor, in foot-candles.
        """
        # INVOKE <id> LightSensor.GetLevelHW
        # -> R:INVOKE <id> <level> LightSensor.GetLevelHW
        return await self.invoke(vid, "LightSensor.GetLevelHW", as_type=Decimal)



"""Interface for querying and controlling anemo (wind) sensors."""

from decimal import Decimal

from .base import Interface


class AnemoSensorInterface(Interface):
    """Interface for querying and controlling anemo (wind) sensors."""

    method_signatures = {
        "AnemoSensor.GetSpeed": Decimal,
        "AnemoSensor.GetSpeedHW": Decimal,
    }

    async def get_speed(self, vid: int) -> Decimal:
        """Get the speed of an anemo sensor, using cached value if available.

        Args:
            vid: The Vantage ID of the anemo sensor.

        Returns:
            The speed of the anemo sensor, in mph.
        """
        # INVOKE <id> AnemoSensor.GetSpeed
        # -> R:INVOKE <id> <speed> AnemoSensor.GetSpeed
        return await self.invoke(vid, "AnemoSensor.GetSpeed", as_type=Decimal)

    async def get_speed_hw(self, vid: int) -> Decimal:
        """Get the speed of an anemo sensor directly from the hardware.

        Args:
            vid: The Vantage ID of the anemo sensor.

        Returns:
            The speed of the anemo sensor, in mph.
        """
        # INVOKE <id> AnemoSensor.GetSpeedHW
        # -> R:INVOKE <id> <speed> AnemoSensor.GetSpeedHW
        return await self.invoke(vid, "AnemoSensor.GetSpeedHW", as_type=Decimal)



"""Somfy URTSI 2 port device."""
from dataclasses import dataclass

from .port_device import PortDevice


@dataclass
class SomfyURTSI2Port(PortDevice):
    """Somfy URTSI 2 port device."""

    class Meta:
        name = "Somfy.URTSI_2_PORT"



"""IConfiguration.GetFilterResults method definition."""

from dataclasses import dataclass, field

from aiovantage.config_client.interfaces.types import ObjectChoice


@dataclass
class GetFilterResults:
    """IConfiguration.GetFilterResults method definition."""

    interface = "IConfiguration"

    @dataclass
    class Params:
        """Method parameters."""

        h_filter: int = field(
            metadata={
                "name": "hFilter",
            }
        )

        count: int = field(
            default=50,
            metadata={
                "name": "Count",
            },
        )

        whole_object: bool = field(
            default=True,
            metadata={
                "name": "WholeObject",
            },
        )

    call: Params | None = field(default=None)
    return_value: list[ObjectChoice] | None = field(
        default=None,
        metadata={
            "name": "Object",
            "wrapper": "return",
        },
    )



"""Client for the Vantage Application Communication Interface (ACI) service.

The ACI service is an XML-based RPC service that Design Center uses to communicate
with Vantage InFusion Controllers. There are a number of "interfaces" exposed, each
with one or more "methods".

This service allows you to query the "configuration" of a Vantage system, for
example fetching a list of all the objects, getting a backup of the Design Center
XML, etc.

The service is exposed on port 2010 (SSL) by default, and on port 2001 (non-SSL) if
this port has been opened by the firewall on the controller.

The service is discoverable via mDNS as `_aci._tcp.local` and/or
`_secure_aci._tcp.local`.
"""

import asyncio
import logging
from ssl import SSLContext
from types import TracebackType
from xml.etree import ElementTree

from typing_extensions import Self
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.handlers import XmlEventHandler
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

from aiovantage.connection import BaseConnection
from aiovantage.errors import ClientResponseError, LoginFailedError, LoginRequiredError

from .interfaces import Call, Method, Return
from .interfaces.introspection import GetSysInfo
from .interfaces.login import Login


class ConfigConnection(BaseConnection):
    """Connection to a Vantage ACI server."""

    default_port = 2001
    default_ssl_port = 2010
    buffer_limit = 2**20


class ConfigClient:
    """Client for the Vantage Application Communication Interface (ACI) service.

    This client handles connecting to the ACI service, authenticating, and the
    serialization/deserialization of XML requests and responses.
    """

    def __init__(
        self,
        host: str,
        username: str | None = None,
        password: str | None = None,
        *,
        ssl: SSLContext | bool = True,
        port: int | None = None,
        conn_timeout: float = 30,
        read_timeout: float = 60,
    ) -> None:
        """Initialize the client."""
        self._connection = ConfigConnection(host, port, ssl, conn_timeout)
        self._username = username
        self._password = password
        self._read_timeout = read_timeout

        self._serializer = XmlSerializer(
            config=SerializerConfig(xml_declaration=False),
        )

        self._parser = XmlParser(
            config=ParserConfig(fail_on_unknown_properties=False),
            handler=XmlEventHandler,
        )
        self._connection_lock = asyncio.Lock()
        self._request_lock = asyncio.Lock()
        self._logger = logging.getLogger(__name__)

    async def __aenter__(self) -> Self:
        """Return context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit context manager."""
        self.close()
        if exc_val:
            raise exc_val

    def close(self) -> None:
        """Close the connection to the ACI service."""
        self._connection.close()

    async def raw_request(
        self,
        interface: str,
        raw_method: str,
        connection: ConfigConnection | None = None,
    ) -> str:
        """Send a raw request to the ACI service and return the raw response.

        Args:
            interface: The interface to send the request to.
            raw_method: The raw XML request to send.
            connection: The connection to use, if not the default.

        Returns:
            The raw XML response.
        """
        # Open the connection if it's closed
        conn = connection or await self.get_connection()

        # Render the method object to XML with xsdata
        request = f"<{interface}>{raw_method}</{interface}>"
        self._logger.debug("Sending request: %s", request)

        # Send the request and read the response
        async with self._request_lock:
            await conn.write(request)
            response = await conn.readuntil(
                f"</{interface}>\n".encode(), timeout=self._read_timeout
            )

        self._logger.debug("Received response: %s", response)

        return response

    async def request(
        self,
        method_cls: type[Method[Call, Return]],
        params: Call | None = None,
        connection: ConfigConnection | None = None,
    ) -> Return | None:
        """Marshall a request, send it to the ACI service, and return a parsed object.

        Args:
            method_cls: The request method class to use.
            params: The parameters to pass to the method.
            connection: The connection to use, if not the default.

        Returns:
            The parsed response object
        """
        # Build the method object
        method = method_cls()
        method.call = params

        # Render the method object to XML with xsdata and send the request
        response = await self.raw_request(
            method.interface, self._serializer.render(method), connection
        )

        # Parse the XML doc
        root = ElementTree.fromstring(response)

        # Response root must match the tag of the request
        if root.tag != method.interface:
            raise ClientResponseError(
                f"Response '{root.tag}' does not match request '{method.interface}'"
            )

        # Responses must contain the method element and a return element
        method_el = root.find(f"./{method_cls.__name__}")
        return_el = root.find(f"./{method_cls.__name__}/return")
        if method_el is None or return_el is None:
            raise ClientResponseError("Response is missing method or return element")

        # Return None if the method has empty return element.
        # This can happen when the client is not logged in, or when the method returns
        # no results, eg. IConfiguration.GetFilterResults.
        if return_el.text is None and len(return_el) == 0:
            return None

        # Parse the method element with xsdata
        method = self._parser.parse(method_el, method_cls)
        if method.return_value is None:
            return None

        return method.return_value

    async def get_connection(self) -> ConfigConnection:
        """Get a connection to the ACI service."""
        async with self._connection_lock:
            if self._connection.closed:
                await self._connection.open()

                # Ensure the connection is authenticated, if required
                if self._username and self._password:
                    # Log in if we have credentials
                    success = await self.request(
                        Login,
                        Login.Params(self._username, self._password),
                        self._connection,
                    )

                    if not success:
                        raise LoginFailedError("Login failed, bad username or password")
                else:
                    # Check if login is required if we don't have credentials
                    sys_info_response = await self.request(
                        GetSysInfo,
                        connection=self._connection,
                    )

                    if sys_info_response is None:
                        raise LoginRequiredError(
                            "Login required, but no credentials were provided"
                        )

                self._logger.info(
                    "Connected to config client at %s:%d",
                    self._connection.host,
                    self._connection.port,
                )

            return self._connection



"""Controller holding and managing Vantage load groups."""

from decimal import Decimal

from typing_extensions import override

from aiovantage.command_client.object_interfaces import LoadInterface
from aiovantage.models import Load, LoadGroup
from aiovantage.query import QuerySet

from .base import BaseController


class LoadGroupsController(BaseController[LoadGroup], LoadInterface):
    """Controller holding and managing Vantage load groups."""

    vantage_types = ("LoadGroup",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("LOAD",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> None:
        """Fetch the state properties of a load."""
        state = {
            "level": await LoadInterface.get_level(self, vid),
        }

        self.update_state(vid, state)

    @override
    def handle_status(self, vid: int, status: str, *args: str) -> None:
        """Handle simple status messages from the event stream."""
        if status != "LOAD":
            return

        # STATUS LOAD
        # -> S:LOAD <id> <level (0-100)>
        state = {
            "level": Decimal(args[0]),
        }

        self.update_state(vid, state)

    def loads(self, vid: int) -> QuerySet[Load]:
        """Return a queryset of all loads in this load group."""
        load_group = self[vid]

        return self._vantage.loads.filter(lambda load: load.id in load_group.load_ids)



"""Controller holding and managing Vantage stations."""

from aiovantage.models import StationObject

from .base import BaseController


class StationsController(BaseController[StationObject]):
    """Controller holding and managing Vantage stations.

    Stations typically represent keypads or remote relays. It is mostly useful
    to know about these devices so we can set up a proper device hierarchy.
    """

    vantage_types = (
        "Dimmer",
        "DualRelayStation",
        "EqCtrl",
        "EqUX",
        "Keypad",
        "ScenePointRelay",
        "Vantage.DmxDaliGateway",
    )
    """The Vantage object types that this controller will fetch."""



"""Fetch all blinds from the Vantage controller, and print out any state changes."""

import argparse
import asyncio
import contextlib
import logging
from typing import Any

from aiovantage import Vantage, VantageEvent
from aiovantage.models import SystemObject

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


def callback(event: VantageEvent, obj: SystemObject, data: dict[str, Any]) -> None:
    """Print out any state changes."""
    if event == VantageEvent.OBJECT_ADDED:
        print(f"[Blind added] '{obj.name}' ({obj.id})")

    elif event == VantageEvent.OBJECT_UPDATED:
        print(f"[Blind updated] '{obj.name}' ({obj.id})")
        for attr in data.get("attrs_changed", []):
            print(f"    {attr} = {getattr(obj, attr)}")


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Subscribe to updates for all blinds
        vantage.blinds.subscribe(callback)

        # Fetch all known blinds from the controller
        await vantage.blinds.initialize()

        # Keep running for a while
        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""AnemoSensor (wind sensor) object."""

from dataclasses import dataclass, field
from decimal import Decimal

from .sensor import Sensor
from .types import Parent


@dataclass
class AnemoSensor(Sensor):
    """AnemoSensor (wind sensor) object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    speed: Decimal | None = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Controllers."""

from .anemo_sensors import AnemoSensorsController
from .areas import AreasController
from .base import BaseController
from .blind_groups import BlindGroupsController
from .blinds import BlindsController
from .buttons import ButtonsController
from .dry_contacts import DryContactsController
from .gmem import GMemController
from .light_sensors import LightSensorsController
from .load_groups import LoadGroupsController
from .loads import LoadsController
from .masters import MastersController
from .modules import ModulesController
from .omni_sensors import OmniSensorsController
from .port_devices import PortDevicesController
from .power_profiles import PowerProfilesController
from .rgb_loads import RGBLoadsController
from .stations import StationsController
from .tasks import TasksController
from .temperature_sensors import TemperatureSensorsController
from .thermostats import ThermostatsController

__all__ = [
    "AnemoSensorsController",
    "AreasController",
    "BaseController",
    "BlindGroupsController",
    "BlindsController",
    "ButtonsController",
    "DryContactsController",
    "GMemController",
    "LightSensorsController",
    "LoadGroupsController",
    "LoadsController",
    "MastersController",
    "ModulesController",
    "OmniSensorsController",
    "PortDevicesController",
    "PowerProfilesController",
    "RGBLoadsController",
    "StationsController",
    "TasksController",
    "TemperatureSensorsController",
    "ThermostatsController",
]



"""IConfiguration.CloseFilter method definition."""

from dataclasses import dataclass, field


@dataclass
class CloseFilter:
    """IConfiguration.CloseFilter method definition."""

    interface = "IConfiguration"

    call: int | None = field(default=None)
    return_value: bool | None = field(
        default=None,
        metadata={
            "name": "return",
        },
    )



"""Print a tree of objects in the Vantage system."""

import argparse
import asyncio
import contextlib
import logging

from aiovantage import Vantage
from aiovantage.models import Area, Load

# Grab connection info from command line arguments
parser = argparse.ArgumentParser(description="aiovantage example")
parser.add_argument("host", help="hostname of Vantage controller")
parser.add_argument("--username", help="username for Vantage controller")
parser.add_argument("--password", help="password for Vantage controller")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


# Some ANSI escape codes for pretty printing
RESET = "\033[0m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
CYAN = "\033[36m"


def colorize(text: str, color: str) -> str:
    """Return text wrapped in the given ANSI color code."""
    return f"{color}{text}{RESET}"


def print_indented(text: str, indent_level: int) -> None:
    """Print text indented by the given number of levels."""
    indent = indent_level * "    "
    print(f"{indent}{text}")


def load_state(load: Load) -> str:
    """Return a string describing the state of a load."""
    if load.level is None:
        return ""

    if load.level == 0:
        return colorize("(OFF)", RED)

    if load.level == 100:
        return colorize("(ON)", GREEN)

    return colorize(f"({load.level}%)", GREEN)


def print_area(vantage: Vantage, area: Area | None, indent: int = 0) -> None:
    """Recursively print an area and all its children."""
    if area is None:
        return

    # Print the area name
    print_indented(colorize(area.name, CYAN), indent)

    # Print loads in this area, if any
    loads = vantage.loads.filter(area_id=area.id)
    if loads:
        print_indented(colorize("Loads", YELLOW), indent + 1)
        for load in loads:
            print_indented(f"- {load.name} {load_state(load)}", indent + 1)
        print()

    # Print RGB loads in this area, if any
    rgb_loads = vantage.rgb_loads.filter(area_id=area.id)
    if rgb_loads:
        print_indented(colorize("RGB Loads", YELLOW), indent + 1)
        for rgb_load in rgb_loads:
            print_indented(f"- {rgb_load.name} ", indent + 1)
        print()

    # Print stations in this area, if any
    stations = vantage.stations.filter(area_id=area.id)
    if stations:
        print_indented(colorize("Stations", YELLOW), indent + 1)
        for station in stations:
            print_indented(f"- {station.name}", indent + 1)
        print()

    # Print dry contacts in this area, if any
    dry_contacts = vantage.dry_contacts.filter(area_id=area.id)
    if dry_contacts:
        print_indented(colorize("Dry Contacts", YELLOW), indent + 1)
        for dry_contact in dry_contacts:
            print_indented(f"- {dry_contact.name}", indent + 1)
        print()

    # Print any child areas
    child_areas = vantage.areas.filter(area_id=area.id)
    if child_areas:
        for child_area in child_areas:
            print_area(vantage, child_area, indent + 1)


async def main() -> None:
    """Run code example."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    # Connect to the Vantage controller
    async with Vantage(args.host, args.username, args.password) as vantage:
        # Preload all the objects we want to dump
        await vantage.initialize()

        # Recursively print the root area and all its children
        print_area(vantage, vantage.areas.get(area_id=0))


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Interface for querying and controlling RGB loads."""

from decimal import Decimal
from enum import IntEnum
from typing import NamedTuple

from .base import Interface


class RGBLoadInterface(Interface):
    """Interface for querying and controlling RGB loads."""

    class RGBChannel(IntEnum):
        """The RGB color channels."""

        Red = 0
        Green = 1
        Blue = 2
        White = 3

    class HSLAttribute(IntEnum):
        """The HSL color attributes."""

        Hue = 0
        Saturation = 1
        Lightness = 2

    class ColorChannelResponse(NamedTuple):
        """A RGB(W) color channel response."""

        value: int
        channel: int

    method_signatures = {
        "RGBLoad.GetRGB": ColorChannelResponse,
        "RGBLoad.GetRGBW": ColorChannelResponse,
        "RGBLoad.GetHSL": ColorChannelResponse,
        "RGBLoad.GetColor": int,
    }

    async def set_rgb(self, vid: int, red: int, green: int, blue: int) -> None:
        """Set the color of an RGB load.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
        """
        # Clamp levels to 0-255
        red = max(min(red, 255), 0)
        green = max(min(green, 255), 0)
        blue = max(min(blue, 255), 0)

        # INVOKE <id> RGBLoad.SetRGB <red> <green> <blue>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGB <red> <green> <blue>
        await self.invoke(vid, "RGBLoad.SetRGB", red, green, blue)

    async def get_rgb(self, vid: int, channel: int) -> int:
        """Get a single RGB color channel of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to get the color of.

        Returns:
            The value of the RGB channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGB <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGB <channel>
        response = await RGBLoadInterface.invoke(
            self, vid, "RGBLoad.GetRGB", channel, as_type=self.ColorChannelResponse
        )
        return response.value

    async def set_hsl(
        self,
        vid: int,
        hue: int,
        saturation: float | Decimal,
        lightness: float | Decimal,
    ) -> None:
        """Set the color of an HSL load.

        Args:
            vid: The Vantage ID of the RGB load.
            hue: The hue value of the color, in degrees (0-360).
            saturation: The saturation value of the color, in percent (0-100).
            lightness: The lightness value of the color, in percent (0-100).
        """
        # Clamp levels to 0-360, 0-100
        hue = max(min(hue, 360), 0)
        saturation = max(min(saturation, 100), 0)
        lightness = max(min(lightness, 100), 0)

        # INVOKE <id> RGBLoad.SetHSL <hue> <saturation> <lightness>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetHSL <hue> <saturation> <lightness>
        await self.invoke(vid, "RGBLoad.SetHSL", hue, saturation, lightness)

    async def get_hsl(self, vid: int, attribute: int) -> int:
        """Get a single HSL color attribute of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            attribute: The attribute to get the value of.

        Returns:
            The value of the HSL attribute, 0-360 for hue, 0-100 for saturation and
            lightness.
        """
        # INVOKE <id> RGBLoad.GetHSL <attribute>
        # -> R:INVOKE <id> <value> RGBLoad.GetHSL <attribute>
        response = await RGBLoadInterface.invoke(
            self, vid, "RGBLoad.GetHSL", attribute, as_type=self.ColorChannelResponse
        )
        return response.value

    async def dissolve_rgb(
        self, vid: int, red: float, green: float, blue: float, rate: float
    ) -> None:
        """Transition the color of an RGB load over a number of seconds.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The new red value of the color, (0-255)
            green: The new green value of the color, (0-255)
            blue: The new blue value of the color, (0-255)
            rate: The number of seconds the transition should take.
        """
        # Clamp levels to 0-255, ensure they're integers
        red = int(max(min(red, 255), 0))
        green = int(max(min(green, 255), 0))
        blue = int(max(min(blue, 255), 0))

        # INVOKE <id> RGBLoad.DissolveRGB <red> <green> <blue> <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveRGB <red> <green> <blue> <rate>
        await self.invoke(vid, "RGBLoad.DissolveRGB", red, green, blue, rate)

    async def dissolve_hsl(
        self, vid: int, hue: float, saturation: float, lightness: float, rate: float
    ) -> None:
        """Transition the color of an HSL load over a number of seconds.

        Args:
            vid: The Vantage ID of the RGB load.
            hue: The new hue value of the color, in degrees (0-360).
            saturation: The new saturation value of the color, in percent (0-100).
            lightness: The new lightness value of the color, in percent (0-100).
            rate: The number of seconds the transition should take.
        """
        # Clamp levels to 0-360, 0-100, ensure they're integers
        hue = int(max(min(hue, 360), 0))
        saturation = int(max(min(saturation, 100), 0))
        lightness = int(max(min(lightness, 100), 0))

        # INVOKE <id> RGBLoad.DissolveHSL <hue> <saturation> <lightness> <rate>
        # -> R:INVOKE <id> <rcode> RGBLoad.DissolveHSL <hue> <saturation> <lightness> <rate>
        await self.invoke(vid, "RGBLoad.DissolveHSL", hue, saturation, lightness, rate)

    async def set_rgb_component(
        self, vid: int, channel: RGBChannel, value: int
    ) -> None:
        """Set a single RGB(W) color channel of a load.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to set the color of.
            value: The value to set the channel to, 0-255.
        """
        # Clamp value to 0-255
        value = max(min(value, 255), 0)

        # INVOKE <id> RGBLoad.SetRGBComponent <channel> <value>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBComponent <channel> <value>
        await self.invoke(vid, "RGBLoad.SetRGBComponent", channel, value)

    async def set_hsl_attribute(
        self, vid: int, attribute: HSLAttribute, value: int
    ) -> None:
        """Set a single HSL color attribute of a load.

        Args:
            vid: The Vantage ID of the RGB load.
            attribute: The attribute to set the value of.
            value: The value to set the attribute to, 0-360 for hue, 0-100 for
                saturation and lightness.
        """
        # Clamp value to 0-360, 0-100
        if attribute == self.HSLAttribute.Hue:
            value = max(min(value, 360), 0)
        else:
            value = max(min(value, 100), 0)

        # INVOKE <id> RGBLoad.SetHSLAttribute <attribute> <value>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetHSLAttribute <attribute> <value>
        await self.invoke(vid, "RGBLoad.SetHSLAttribute", attribute, value)

    async def get_color(self, vid: int) -> int:
        """Get the RGB/RGBW color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The RGB(W) value of the color as a packed big-endian integer.
        """
        # INVOKE <id> RGBLoad.GetColor
        # -> R:INVOKE <id> <color> RGBLoad.GetColor
        return await self.invoke(vid, "RGBLoad.GetColor", as_type=int)

    async def set_rgbw(
        self, vid: int, red: float, green: float, blue: float, white: float
    ) -> None:
        """Set the color of an RGBW load.

        Args:
            vid: The Vantage ID of the RGB load.
            red: The red value of the color, (0-255)
            green: The green value of the color, (0-255)
            blue: The blue value of the color, (0-255)
            white: The white value of the color, (0-255)
        """
        # Clamp levels to 0-255
        red = int(max(min(red, 255), 0))
        green = int(max(min(green, 255), 0))
        blue = int(max(min(blue, 255), 0))
        white = int(max(min(white, 255), 0))

        # INVOKE <id> RGBLoad.SetRGBW <red> <green> <blue> <white>
        # -> R:INVOKE <id> <rcode> RGBLoad.SetRGBW <red> <green> <blue> <white>
        await self.invoke(vid, "RGBLoad.SetRGBW", red, green, blue, white)

    async def get_rgbw(self, vid: int, channel: int) -> int:
        """Get a single RGBW color channel of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.
            channel: The channel to get the color of.

        Returns:
            The value of the RGBW channel, 0-255.
        """
        # INVOKE <id> RGBLoad.GetRGBW <channel>
        # -> R:INVOKE <id> <value> RGBLoad.GetRGBW <channel>
        response = await self.invoke(
            vid, "RGBLoad.GetRGBW", channel, as_type=self.ColorChannelResponse
        )
        return response.value

    # Additional convenience methods, not part of the Vantage API
    async def get_rgb_color(self, vid: int) -> tuple[int, ...]:
        """Get the RGB color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the RGB color as a tuple of (red, green, blue).
        """
        return tuple([await self.get_rgb(vid, attr) for attr in range(3)])

    async def get_rgbw_color(self, vid: int) -> tuple[int, ...]:
        """Get the RGBW color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the RGBW color as a tuple of (red, green, blue, white).
        """
        return tuple([await self.get_rgbw(vid, chan) for chan in range(4)])

    async def get_hsl_color(self, vid: int) -> tuple[int, ...]:
        """Get the HSL color of a load from the controller.

        Args:
            vid: The Vantage ID of the RGB load.

        Returns:
            The value of the HSL color as a tuple of (hue, saturation, lightness).
        """
        return tuple([await self.get_hsl(vid, attr) for attr in range(3)])



"""Interface for keypad speakers."""

from decimal import Decimal
from enum import IntEnum

from .base import Interface


class SounderInterface(Interface):
    """Interface for keypad speakers."""

    class Status(IntEnum):
        """Sounder status."""

        On = 0
        Off = 1

    method_signatures = {
        "Sounder.GetFrequency": Decimal,
        "Sounder.GetFrequencyHW": Decimal,
        "Sounder.GetStatus": Status,
        "Sounder.GetStatusHW": Status,
    }

    async def get_frequency(self, vid: int) -> Decimal:
        """Get the frequency of the keypad speaker, using cached value if available.

        Args:
            vid: The Vantage ID of the keypad.

        Returns:
            The frequency of the keypad speaker in Hz.
        """
        # INVOKE <id> Sounder.GetFrequency
        # -> R:INVOKE <id> <frequency> Sounder.GetFrequency
        return await self.invoke(vid, "Sounder.GetFrequency", as_type=Decimal)

    async def get_frequency_hw(self, vid: int) -> Decimal:
        """Get the frequency of the keypad speaker directly from the hardware.

        Args:
            vid: The Vantage ID of the keypad.

        Returns:
            The frequency of the keypad speaker in Hz.
        """
        # INVOKE <id> Sounder.GetFrequencyHW
        # -> R:INVOKE <id> <frequency> Sounder.GetFrequency
        return await self.invoke(vid, "Sounder.GetFrequencyHW", as_type=Decimal)

    async def set_frequency(self, vid: int, frequency: float | Decimal) -> None:
        """Set the frequency of the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
            frequency: The frequency to set the sounder to, in Hz.
        """
        # INVOKE <id> Sounder.SetFrequency <frequency>
        # -> R:INVOKE <id> <rcode> Sounder.SetFrequency
        await self.invoke(vid, "Sounder.SetFrequency", frequency)

    async def get_status(self, vid: int) -> Status:
        """Get the status of the keypad speaker, using cached value if available.

        Args:
            vid: The Vantage ID of the keypad.

        Returns:
            The status of the keypad speaker.
        """
        # INVOKE <id> Sounder.GetStatus
        # -> R:INVOKE <id> <status (0/1)> Sounder.GetStatus
        return await self.invoke(vid, "Sounder.GetStatus", as_type=self.Status)

    async def get_status_hw(self, vid: int) -> Status:
        """Get the status of the keypad speaker directly from the hardware.

        Args:
            vid: The Vantage ID of the keypad.

        Returns:
            The status of the keypad speaker.
        """
        # INVOKE <id> Sounder.GetStatusHW
        # -> R:INVOKE <id> <status (0/1)> Sounder.GetStatus
        return await self.invoke(vid, "Sounder.GetStatusHW", as_type=self.Status)

    async def set_status(self, vid: int, status: Status) -> None:
        """Set the status of the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
            status: The status to set the keypad speaker to.
        """
        # INVOKE <id> Sounder.SetStatus <status (0/1/On/Off)>
        # -> R:INVOKE <id> <rcode> Sounder.SetStatus
        await self.invoke(vid, "Sounder.SetStatus", status)

    async def play_fx(
        self, vid: int, effect: int, duration: float = 0, volume: float = 0
    ) -> None:
        """Play a sound effect on the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
            effect: The effect to play.
            duration: The duration to play the FX for, in seconds, 0 for default.
            volume: The volume to play the FX at, as a percentage, 0 for default.
        """
        # INVOKE <id> Sounder.PlayFX <fx> <duration> <volume>
        # -> R:INVOKE <id> <rcode> Sounder.PlayFX
        await self.invoke(vid, "Sounder.PlayFX", effect, duration, volume)

    # Additional convenience methods, not part of the Vantage API
    async def turn_on(self, vid: int) -> None:
        """Turn on the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
        """
        await self.set_status(vid, self.Status.On)

    async def turn_off(self, vid: int) -> None:
        """Turn off the keypad speaker.

        Args:
            vid: The Vantage ID of the keypad.
        """
        await self.set_status(vid, self.Status.Off)
